# pragma: no cover
import sys
import asyncio

import synapse.lib.jsonstor as s_jsonstor

if __name__ == '__main__':  # pragma: no cover
    asyncio.run(s_jsonstor.JsonStorCell.execmain(sys.argv[1:]))



import synapse.lib.module as s_module


class MediaModule(s_module.CoreModule):

    def getModelDefs(self):
        name = 'media'

        ctors = ()

        forms = (
            ('media:news', {}, (
                ('url', ('inet:url', {}), {
                    'doc': 'The (optional) URL where the news was published.',
                    'ex': 'http://cnn.com/news/mars-lander.html',
                }),
                ('url:fqdn', ('inet:fqdn', {}), {
                    'doc': 'The FQDN within the news URL.',
                    'ex': 'cnn.com',
                }),
                ('file', ('file:bytes', {}), {
                    'doc': 'The (optional) file blob containing or published as the news.',
                }),
                ('title', ('str', {'lower': True}), {
                    'doc': 'Title/Headline for the news.',
                    'ex': 'mars lander reaches mars',
                    'disp': {'hint': 'text'},
                }),
                ('summary', ('str', {}), {
                    'doc': 'A brief summary of the news item.',
                    'ex': 'lorum ipsum',
                    'disp': {'hint': 'text'},
                }),
                ('published', ('time', {}), {
                    'doc': 'The date the news item was published.',
                    'ex': '20161201180433',
                }),
                ('org', ('ou:alias', {}), {
                    'doc': 'The org alias which published the news.',
                    'ex': 'microsoft',
                }),
                ('author', ('ps:name', {}), {
                    'doc': 'The free-form author of the news.',
                    'deprecated': True,
                    'ex': 'stark,anthony'
                }),
                ('authors', ('array', {'type': 'ps:contact', 'split': ',', 'uniq': True, 'sorted': True}), {
                    'doc': 'An array of authors of the news item.',
                }),
                ('rss:feed', ('inet:url', {}), {
                    'doc': 'The RSS feed that published the news.',
                }),
            )),
        )

        types = (
            ('media:news', ('guid', {}), {
                'doc': 'A GUID for a news article or report.'
            }),
        )

        modldef = (name, {
            'ctors': ctors,
            'forms': forms,
            'types': types,
        })
        return (modldef, )



import asyncio

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.stormtypes as s_stormtypes

from synapse.lib.stormtypes import tostr

@s_stormtypes.registry.registerType
class ProjectEpic(s_stormtypes.Prim):
    '''
    Implements the Storm API for a ProjectEpic
    '''
    _storm_locals = (
        {'name': 'name', 'desc': 'The name of the Epic', 'type': 'str'},
    )
    _storm_typename = 'storm:project:epic'

    def __init__(self, proj, node):
        s_stormtypes.Prim.__init__(self, None)
        self.proj = proj
        self.node = node
        self.locls.update(self.getObjLocals())
        self.stors.update({
            'name': self._setEpicName,
        })

    def getObjLocals(self):
        return {
            'name': self.node.get('name'),
        }

    async def value(self):
        return self.node.ndef[1]

    async def nodes(self):
        yield self.node

    async def _setEpicName(self, valu):
        self.proj.confirm(('project', 'epic', 'set', 'name'))
        name = await tostr(valu, noneok=True)
        if name is None:
            await self.node.pop('name')
        else:
            await self.node.set('name', name)

@s_stormtypes.registry.registerType
class ProjectEpics(s_stormtypes.Prim):
    '''
    Implements the Storm API for ProjectEpics objects, which are collections of ProjectEpic
    objects associated with a particular Project
    '''
    _storm_locals = (
        {'name': 'get', 'desc': 'Get an epic by name.',
         'type': {'type': 'function', '_funcname': '_getProjEpic',
                  'args': (
                      {'name': 'name', 'type': 'str', 'desc': 'The name (or iden) of the ProjectEpic to get.'},
                  ),
                  'returns': {'type': 'storm:project:epic', 'desc': 'The `storm:project:epic` object', }}},
        {'name': 'add', 'desc': 'Add an epic.',
         'type': {'type': 'function', '_funcname': '_addProjEpic',
                  'args': (
                      {'name': 'name', 'type': 'str', 'desc': 'The name for the new ProjectEpic.'},
                  ),
                  'returns': {'type': 'storm:project:epic', 'desc': 'The newly created `storm:project:epic` object', }}},
        {'name': 'del', 'desc': 'Delete an epic by name.',
         'type': {'type': 'function', '_funcname': '_delProjEpic',
                  'args': (
                      {'name': 'name', 'type': 'str', 'desc': 'The name of the ProjectEpic to delete.'},
                  ),
                  'returns': {'type': 'boolean', 'desc': 'True if the ProjectEpic can be found and deleted, otherwise False', }}}
    )
    _storm_typename = 'storm:project:epics'

    def __init__(self, proj):
        s_stormtypes.Prim.__init__(self, None)
        self.proj = proj
        self.locls.update(self.getObjLocals())

    def getObjLocals(self):
        return {
            'get': self._getProjEpic,
            'add': self._addProjEpic,
            'del': self._delProjEpic,
        }

    async def _getProjEpic(self, name):
        return await self.proj._getProjEpic(name)

    async def _delProjEpic(self, name):

        self.proj.confirm(('project', 'epic', 'del'))
        epic = await self.proj._getProjEpic(name)
        if epic is None:
            return False

        nodeedits = []
        async for tick in self.proj.runt.snap.nodesByPropValu('proj:ticket:epic', '=', epic.node.ndef[1]):
            nodeedits.append(
                (tick.buid, 'proj:ticket', await tick._getPropDelEdits('epic'))
            )
            await asyncio.sleep(0)

        await self.proj.runt.snap.applyNodeEdits(nodeedits)
        await epic.node.delete()
        return True

    async def _addProjEpic(self, name):
        self.proj.confirm(('project', 'epic', 'add'))
        tick = s_common.now()
        props = {
            'name': await tostr(name),
            'created': tick,
            'creator': self.proj.runt.user.iden,
            'project': self.proj.node.ndef[1],
        }
        node = await self.proj.runt.snap.addNode('proj:epic', '*', props=props)
        return ProjectEpic(self.proj, node)

    async def iter(self):
        opts = {'vars': {'proj': self.proj.node.ndef[1]}}
        async for node, path in self.proj.runt.storm('proj:epic:project=$proj', opts=opts):
            yield ProjectEpic(self.proj, node)

@s_stormtypes.registry.registerType
class ProjectTicketComment(s_stormtypes.Prim):
    '''
    Implements the Storm API for a ProjectTicketComment
    '''
    _storm_locals = (
        {'name': 'text', 'desc': 'The comment text.',
         'type': 'str'},
        {'name': 'del', 'desc': 'Delete the comment.',
         'type': {'type': 'function', '_funcname': '_delTicketComment',
                  'returns': {'type': 'boolean', 'desc': 'True if the ProjectTicketComment was deleted'}}},
    )

    _storm_typename = 'storm:project:ticket:comment'

    def __init__(self, ticket, node):
        s_stormtypes.Prim.__init__(self, None)
        self.node = node
        self.proj = ticket.proj
        self.ticket = ticket
        self.locls.update(self.getObjLocals())
        self.stors.update({
            'text': self._setCommentText,
        })

    def getObjLocals(self):
        return {
            'text': self.node.get('text'),
            'del': self._delTicketComment,
        }

    @s_stormtypes.stormfunc(readonly=True)
    async def value(self):
        if self.node is None:
            raise s_exc.StormRuntimeError(mesg='Comment has been deleted')
        return self.node.ndef[1]

    async def nodes(self):
        yield self.node

    async def _setCommentText(self, valu):

        if self.node is None:
            raise s_exc.StormRuntimeError(mesg='Comment has been deleted')

        useriden = self.proj.runt.user.iden
        if useriden != self.node.get('creator'):
            raise s_exc.AuthDeny(mesg='Comment was created by a different user')

        strvalu = await tostr(valu)
        await self.node.set('text', strvalu)
        await self.node.set('updated', s_common.now())

    async def _delTicketComment(self):

        if self.node is None:
            raise s_exc.StormRuntimeError(mesg='Comment has been deleted')

        if self.node.get('creator') != self.proj.runt.user.iden:
            self.proj.confirm(('project', 'comment', 'del'))

        await self.node.delete()
        self.node = None
        return True

@s_stormtypes.registry.registerType
class ProjectTicketComments(s_stormtypes.Prim):
    '''
    Implements the Storm API for ProjectTicketComments objects, which are collections of comments
    associated with a ticket.
    '''
    _storm_locals = (
        {'name': 'get', 'desc': 'Get a ticket comment by guid.',
         'type': {'type': 'function', '_funcname': '_getTicketComment',
                  'args': (
                      {'name': 'guid', 'type': 'str', 'desc': 'The guid of the ProjectTicketComment to get.'},
                  ),
                  'returns': {'type': 'storm:project:ticket:comment',
                              'desc': 'The `storm:project:ticket:comment` object', }}},
        {'name': 'add', 'desc': 'Add a comment to the ticket.',
         'type': {'type': 'function', '_funcname': '_addTicketComment',
                  'args': (
                      {'name': 'text', 'type': 'str', 'desc': 'The text for the new ProjectTicketComment.'},
                  ),
                  'returns': {'type': 'storm:project:ticket:comment',
                              'desc': 'The newly created `storm:project:ticketcomment` object', }}},
    )

    _storm_typename = 'storm:project:ticket:comments'

    def __init__(self, ticket):
        s_stormtypes.Prim.__init__(self, None)
        self.proj = ticket.proj
        self.ticket = ticket
        self.locls.update(self.getObjLocals())

    def getObjLocals(self):
        return {
            'add': self._addTicketComment,
            'get': self._getTicketComment,
        }

    async def _addTicketComment(self, text):
        self.proj.confirm(('project', 'comment', 'add'))
        tick = s_common.now()
        props = {
            'text': await tostr(text),
            'ticket': self.ticket.node.ndef[1],
            'created': tick,
            'updated': tick,
            'creator': self.proj.runt.user.iden,
        }
        node = await self.proj.runt.snap.addNode('proj:comment', '*', props=props)
        return ProjectTicketComment(self.ticket, node)

    @s_stormtypes.stormfunc(readonly=True)
    async def _getTicketComment(self, guid):

        async def filt(node):
            return node.get('ticket') == self.ticket.node.ndef[1]

        guid = await tostr(guid)

        node = await self.proj.runt.getOneNode('proj:comment', guid, filt=filt, cmpr='=')
        if node is not None:
            return ProjectTicketComment(self.ticket, node)

        return None

    @s_stormtypes.stormfunc(readonly=True)
    async def iter(self):
        opts = {'vars': {'ticket': self.ticket.node.ndef[1]}}
        async for node, path in self.proj.runt.storm('proj:comment:ticket=$ticket', opts=opts):
            yield ProjectTicketComment(self.ticket, node)

@s_stormtypes.registry.registerType
class ProjectTicket(s_stormtypes.Prim):
    '''
    Implements the Storm API for a ProjectTicket
    '''
    _storm_locals = (
        {'name': 'name', 'desc': 'The name of the ticket',
         'type': 'str'},
        {'name': 'desc', 'desc': 'A description of the ticket',
         'type': 'str'},
        {'name': 'priority', 'desc': 'An integer value from the enums [0, 10, 20, 30, 40, 50] of the priority of the ticket',
         'type': 'int'},
        {'name': 'comments', 'desc': 'A ``storm:project:ticket:comments`` object that contains comments associated with the given ticket.',
         'type': 'storm:project:ticket:comments'},
    )

    _storm_typename = 'storm:project:ticket'

    def __init__(self, proj, node):
        s_stormtypes.Prim.__init__(self, None)
        self.proj = proj
        self.node = node
        self.locls.update(self.getObjLocals())
        self.ctors.update({
            'comments': self._ctorTicketComments,
        })
        self.stors.update({
            'name': self._setName,
            'desc': self._setDesc,
            'epic': self._setEpic,
            'status': self._setStatus,
            'sprint': self._setSprint,
            'assignee': self._setAssignee,
            'priority': self._setPriority,
        })

    def getObjLocals(self):
        return {
            'name': self.node.get('name'),
            'desc': self.node.get('desc'),
            'priority': self.node.get('priority'),
        }

    @s_stormtypes.stormfunc(readonly=True)
    async def value(self):
        return self.node.ndef[1]

    async def nodes(self):
        yield self.node

    async def _setName(self, valu):

        useriden = self.proj.runt.user.iden
        if useriden != self.node.get('creator'):
            self.proj.confirm(('project', 'ticket', 'set', 'name'))

        strvalu = await tostr(valu, noneok=True)
        if strvalu is None:
            await self.node.pop('name')
        else:
            await self.node.set('name', strvalu)
        await self.node.set('updated', s_common.now())

    async def _setDesc(self, valu):

        useriden = self.proj.runt.user.iden
        if useriden != self.node.get('creator'):
            self.proj.confirm(('project', 'ticket', 'set', 'desc'))

        strvalu = await tostr(valu, noneok=True)
        if strvalu is None:
            await self.node.pop('desc')
        else:
            await self.node.set('desc', strvalu)
        await self.node.set('updated', s_common.now())

    async def _setEpic(self, valu):

        useriden = self.proj.runt.user.iden
        if useriden != self.node.get('creator'):
            self.proj.confirm(('project', 'ticket', 'set', 'epic'))

        strvalu = await tostr(valu, noneok=True)
        if strvalu is None:
            await self.node.pop('epic')
            await self.node.set('updated', s_common.now())
            return

        epic = await self.proj._getProjEpic(strvalu)
        if epic is None:
            mesg = 'No epic found by that name/iden.'
            raise s_exc.NoSuchName(mesg=mesg)

        await self.node.set('epic', epic.node.ndef[1])
        await self.node.set('updated', s_common.now())

    async def _setStatus(self, valu):

        useriden = self.proj.runt.user.iden
        if useriden != self.node.get('assignee'):
            self.proj.confirm(('project', 'ticket', 'set', 'status'))

        strvalu = await tostr(valu)
        await self.node.set('status', strvalu)
        await self.node.set('updated', s_common.now())

    async def _setPriority(self, valu):

        self.proj.confirm(('project', 'ticket', 'set', 'priority'))

        strvalu = await tostr(valu)
        await self.node.set('priority', strvalu)
        await self.node.set('updated', s_common.now())

    async def _setAssignee(self, valu):

        self.proj.confirm(('project', 'ticket', 'set', 'assignee'))

        strvalu = await tostr(valu, noneok=True)

        if strvalu is None:
            await self.node.pop('assignee')
            await self.node.set('updated', s_common.now())
            return

        udef = await self.proj.runt.snap.core.getUserDefByName(strvalu)
        if udef is None:
            mesg = f'No user found by the name {strvalu}'
            raise s_exc.NoSuchUser(mesg=mesg)
        await self.node.set('assignee', udef['iden'])
        await self.node.set('updated', s_common.now())

    async def _setSprint(self, valu):

        self.proj.confirm(('project', 'ticket', 'set', 'sprint'))

        strvalu = await tostr(valu, noneok=True)

        if strvalu is None:
            await self.node.pop('sprint')
            await self.node.set('updated', s_common.now())
            return

        sprint = await self.proj._getProjSprint(strvalu)
        if sprint is None:
            mesg = f'No sprint found by that name/iden ({strvalu}).'
            raise s_exc.NoSuchName(mesg=mesg)

        await self.node.set('sprint', sprint.node.ndef[1])
        await self.node.set('updated', s_common.now())

    def _ctorTicketComments(self, path=None):
        return ProjectTicketComments(self)

@s_stormtypes.registry.registerType
class ProjectTickets(s_stormtypes.Prim):
    '''
    Implements the Storm API for ProjectTickets objects, which are collections of tickets
    associated with a project
    '''

    _storm_typename = 'storm:project:tickets'
    _storm_locals = (
        {'name': 'get', 'desc': 'Get a ticket by name.',
         'type': {'type': 'function', '_funcname': '_getProjTicket',
                  'args': (
                      {'name': 'name', 'type': 'str', 'desc': 'The name (or iden) of the ProjectTicket to get.'},
                  ),
                  'returns': {'type': 'storm:project:ticket', 'desc': 'The `storm:project:ticket` object', }}},
        {'name': 'add', 'desc': 'Add a ticket.',
         'type': {'type': 'function', '_funcname': '_addProjTicket',
                  'args': (
                      {'name': 'name', 'type': 'str', 'desc': 'The name for the new ProjectTicket.'},
                      {'name': 'desc', 'type': 'str', 'desc': 'A description of the new ticket', 'default': ''},
                  ),
                  'returns': {'type': 'storm:project:ticket', 'desc': 'The newly created `storm:project:ticket` object', }}},
        {'name': 'del', 'desc': 'Delete a sprint by name.',
         'type': {'type': 'function', '_funcname': '_delProjTicket',
                  'args': (
                      {'name': 'name', 'type': 'str', 'desc': 'The name of the ProjectTicket to delete.'},
                  ),
                  'returns': {'type': 'boolean', 'desc': 'True if the ProjectTicket can be found and deleted, otherwise False', }}}
    )

    def __init__(self, proj):
        s_stormtypes.Prim.__init__(self, None)
        self.proj = proj
        self.locls.update(self.getObjLocals())

    def getObjLocals(self):
        return {
            'get': self._getProjTicket,
            'add': self._addProjTicket,
            'del': self._delProjTicket,
        }

    @s_stormtypes.stormfunc(readonly=True)
    async def _getProjTicket(self, name):

        async def filt(node):
            return node.get('project') == self.proj.node.ndef[1]

        name = await tostr(name)

        node = await self.proj.runt.getOneNode('proj:ticket:name', name, filt=filt, cmpr='^=')
        if node is not None:
            return ProjectTicket(self.proj, node)

        node = await self.proj.runt.getOneNode('proj:ticket', name, filt=filt, cmpr='^=')
        if node is not None:
            return ProjectTicket(self.proj, node)

        return None

    async def _delProjTicket(self, name):

        tick = await self._getProjTicket(name)
        if tick is None:
            return False

        if tick.node.get('creator') != self.proj.runt.user.iden:
            self.proj.confirm(('project', 'ticket', 'del'))

        # cascade delete comments
        async for node in self.proj.runt.snap.nodesByPropValu('proj:comment:ticket', '=', tick.node.ndef[1]):
            await node.delete()

        await tick.node.delete()
        return True

    async def _addProjTicket(self, name, desc=''):
        self.proj.confirm(('project', 'ticket', 'add'))
        tick = s_common.now()
        props = {
            'name': await tostr(name),
            'desc': await tostr(desc),
            'status': 0,
            'priority': 0,
            'created': tick,
            'updated': tick,
            'creator': self.proj.runt.user.iden,
            'project': self.proj.node.ndef[1],
        }
        node = await self.proj.runt.snap.addNode('proj:ticket', '*', props=props)
        return ProjectTicket(self.proj, node)

    @s_stormtypes.stormfunc(readonly=True)
    async def iter(self):
        opts = {'vars': {'proj': self.proj.node.ndef[1]}}
        async for node, path in self.proj.runt.storm('proj:ticket:project=$proj', opts=opts):
            yield ProjectTicket(self.proj, node)

@s_stormtypes.registry.registerType
class ProjectSprint(s_stormtypes.Prim):
    '''
    Implements the Storm API for a ProjectSprint
    '''

    _storm_locals = (
        {'name': 'name', 'desc': 'The name of the sprint',
         'type': 'str'},
        {'name': 'desc', 'desc': 'A description of the sprint',
         'type': 'str'},
        {'name': 'tickets', 'desc': 'Yields out the tickets associated with the given sprint (no call needed)',
         'type': 'generator'},
    )

    _storm_typename = 'storm:project:sprint'

    def __init__(self, proj, node):
        s_stormtypes.Prim.__init__(self, None)
        self.proj = proj
        self.node = node
        self.locls.update(self.getObjLocals())
        self.ctors.update({
            'tickets': self._getSprintTickets,
        })
        self.stors.update({
            'name': self._setSprintName,
            'desc': self._setSprintDesc,
            'status': self._setSprintStatus,
        })

    async def _setSprintStatus(self, valu):
        self.proj.confirm(('project', 'sprint', 'set', 'status'))
        valu = await tostr(valu, noneok=True)
        if valu is None:
            await self.node.pop('status')
        else:
            await self.node.set('status', valu)

    async def _setSprintDesc(self, valu):
        self.proj.confirm(('project', 'sprint', 'set', 'desc'))
        valu = await tostr(valu, noneok=True)
        if valu is None:
            await self.node.pop('desc')
        else:
            await self.node.set('desc', valu)

    async def _setSprintName(self, valu):

        self.proj.confirm(('project', 'sprint', 'set', 'name'))
        valu = await tostr(valu, noneok=True)
        if valu is None:
            await self.node.pop('name')
        else:
            await self.node.set('name', valu)

    async def _getSprintTickets(self, path=None):
        async for node in self.proj.runt.snap.nodesByPropValu('proj:ticket:sprint', '=', self.node.ndef[1]):
            yield ProjectTicket(self.proj, node)

    def getObjLocals(self):
        return {
            'name': self.node.get('name'),
            'desc': self.node.get('desc'),
            'status': self.node.get('status'),
        }

    @s_stormtypes.stormfunc(readonly=True)
    async def value(self):
        return self.node.ndef[1]

    async def nodes(self):
        yield self.node

@s_stormtypes.registry.registerType
class ProjectSprints(s_stormtypes.Prim):
    '''
    Implements the Storm API for ProjectSprints objects, which are collections of sprints
    associated with a single project
    '''
    _storm_locals = (
        {'name': 'get', 'desc': 'Get a sprint by name.',
         'type': {'type': 'function', '_funcname': '_getProjSprint',
                  'args': (
                      {'name': 'name', 'type': 'str', 'desc': 'The name (or iden) of the ProjectSprint to get.'},
                  ),
                  'returns': {'type': 'storm:project:sprint', 'desc': 'The `storm:project:sprint` object', }}},
        {'name': 'add', 'desc': 'Add a sprint.',
         'type': {'type': 'function', '_funcname': '_addProjSprint',
                  'args': (
                      {'name': 'name', 'type': 'str', 'desc': 'The name for the new ProjectSprint.'},
                      {'name': 'period', 'type': 'ival', 'desc': 'The time interval the ProjectSprint runs for',
                       'default': None},
                  ),
                  'returns': {'type': 'storm:project:sprint', 'desc': 'The newly created `storm:project:sprint` object', }}},
        {'name': 'del', 'desc': 'Delete a sprint by name.',
         'type': {'type': 'function', '_funcname': '_delProjSprint',
                  'args': (
                      {'name': 'name', 'type': 'str', 'desc': 'The name of the Sprint to delete.'},
                  ),
                  'returns': {'type': 'boolean', 'desc': 'True if the ProjectSprint can be found and deleted, otherwise False', }}}
    )

    _storm_typename = 'storm:project:sprints'

    def __init__(self, proj):
        s_stormtypes.Prim.__init__(self, None)
        self.proj = proj
        self.locls.update(self.getObjLocals())

    def getObjLocals(self):
        return {
            'get': self._getProjSprint,
            'add': self._addProjSprint,
            'del': self._delProjSprint,
        }

    async def _getProjSprint(self, name):
        return await self.proj._getProjSprint(name)

    async def _addProjSprint(self, name, period=None):

        self.proj.confirm(('project', 'sprint', 'add'))

        props = {
            'name': await tostr(name),
            'created': s_common.now(),
            'project': self.proj.node.ndef[1],
            'creator': self.proj.runt.snap.user.iden,
            'status': 'planned',
        }

        if period is not None:
            props['period'] = period

        node = await self.proj.runt.snap.addNode('proj:sprint', '*', props=props)
        return ProjectSprint(self.proj, node)

    async def _delProjSprint(self, name):

        self.proj.confirm(('project', 'sprint', 'del'))
        sprint = await self.proj._getProjSprint(name)
        if sprint is None:
            return False

        sprintiden = sprint.node.ndef[1]

        nodeedits = []
        async for tick in self.proj.runt.snap.nodesByPropValu('proj:ticket:sprint', '=', sprintiden):
            nodeedits.append(
                (tick.buid, 'proj:ticket', await tick._getPropDelEdits('sprint'))
            )
            await asyncio.sleep(0)

        await self.proj.runt.snap.applyNodeEdits(nodeedits)
        await sprint.node.delete()
        return True

    @s_stormtypes.stormfunc(readonly=True)
    async def iter(self):
        opts = {'vars': {'proj': self.proj.node.ndef[1]}}
        async for node, path in self.proj.runt.storm('proj:sprint:project=$proj', opts=opts):
            yield ProjectSprint(self.proj, node)

@s_stormtypes.registry.registerType
class Project(s_stormtypes.Prim):
    '''
    Implements the Storm API for Project objects, which are used for managing a scrum style project in the Cortex
    '''
    _storm_locals = (
        {'name': 'name', 'desc': 'The name of the project', 'type': 'str'},
        {'name': 'epics', 'desc': 'A `storm:project:epics` object that contains the epics associated with the given project.',
         'type': 'storm:project:epics'},
        {'name': 'sprints', 'desc': 'A `storm:project:sprints` object that contains the sprints associated with the given project.',
         'type': 'storm:project:sprints'},
        {'name': 'tickets', 'desc': 'A `storm:project:tickets` object that contains the tickets associated with the given project.',
         'type': 'storm:project:tickets'},
    )

    _storm_typename = 'storm:project'

    def __init__(self, runt, node, path=None):
        s_stormtypes.Prim.__init__(self, None)
        self.node = node
        self.runt = runt
        self.locls.update(self.getObjLocals())
        self.ctors.update({
            'epics': self._ctorProjEpics,
            'sprints': self._ctorProjSprints,
            'tickets': self._ctorProjTickets,
        })
        self.stors.update({
            'name': self._setName,
        })

    def _ctorProjEpics(self, path=None):
        return ProjectEpics(self)

    def _ctorProjSprints(self, path=None):
        return ProjectSprints(self)

    def _ctorProjTickets(self, path=None):
        return ProjectTickets(self)

    def confirm(self, perm):
        gateiden = self.node.ndef[1]
        # bypass runt.confirm() here to avoid asroot
        return self.runt.user.confirm(perm, gateiden=gateiden)

    def getObjLocals(self):
        return {
            'name': self.node.get('name'),
        }

    async def _setName(self, valu):
        self.confirm(('project', 'set', 'name'))
        await self.node.set('name', await tostr(valu))

    @s_stormtypes.stormfunc(readonly=True)
    def value(self):
        return self.node.ndef[1]

    async def nodes(self):
        yield self.node

    async def _getProjEpic(self, name):

        async def filt(node):
            return node.get('project') == self.node.ndef[1]

        name = await tostr(name)

        node = await self.runt.getOneNode('proj:epic:name', name, filt=filt, cmpr='^=')
        if node is not None:
            return ProjectEpic(self, node)

        node = await self.runt.getOneNode('proj:epic', name, filt=filt, cmpr='^=')
        if node is not None:
            return ProjectEpic(self, node)

    async def _getProjSprint(self, name):

        async def filt(node):
            return node.get('project') == self.node.ndef[1]

        name = await tostr(name)

        node = await self.runt.getOneNode('proj:sprint:name', name, filt=filt, cmpr='^=')
        if node is not None:
            return ProjectSprint(self, node)

        node = await self.runt.getOneNode('proj:sprint', name, filt=filt, cmpr='^=')
        if node is not None:
            return ProjectSprint(self, node)

        return None

@s_stormtypes.registry.registerLib
class LibProjects(s_stormtypes.Lib):
    '''
    A Storm Library for interacting with Projects in the Cortex.
    '''
    _storm_locals = (
        {'name': 'get', 'desc': 'Retrieve a project by name',
         'type': {'type': 'function', '_funcname': '_funcProjGet',
                  'args': (
                      {'name': 'name', 'type': 'str', 'desc': 'The name of the Project to get'},
                  ),
                  'returns': {'type': 'storm:project',
                              'desc': 'The `storm:project `object, if it exists, otherwise null'}}},

        {'name': 'add', 'desc': 'Add a new project',
         'type': {'type': 'function', '_funcname': '_funcProjAdd',
                  'args': (
                      {'name': 'name', 'type': 'str', 'desc': 'The name of the Project to add'},
                      {'name': 'desc', 'type': 'str', 'desc': 'A description of the overall project', 'default': ''},
                  ),
                  'returns': {'type': 'storm:project', 'desc': 'The newly created `storm:project` object'}}},
        {'name': 'del', 'desc': 'Delete an existing project',
         'type': {'type': 'function', '_funcname': '_funcProjDel',
                  'args': (
                      {'name': 'name', 'type': 'str', 'desc': 'The name of the Project to delete'},
                  ),
                  'returns': {'type': 'boolean',
                              'desc': 'True if the project exists and gets deleted, otherwise False'}}},
    )
    _storm_lib_path = ('projects',)

    def getObjLocals(self):
        return {
            'get': self._funcProjGet,
            'add': self._funcProjAdd,
            'del': self._funcProjDel,
        }

    async def iter(self):
        async for node, path in self.runt.storm('proj:project'):
            yield Project(self.runt, node)

    async def _funcProjDel(self, name):
        gateiden = self.runt.snap.view.iden
        # do not use self.runt.confirm() to avoid asroot
        self.runt.user.confirm(('project', 'del'), gateiden=gateiden)
        proj = await self._funcProjGet(name)
        if proj is None:
            return False

        await proj.node.delete()
        return True

    async def _funcProjGet(self, name):

        name = await tostr(name)

        node = await self.runt.getOneNode('proj:project:name', name, cmpr='^=')
        if node is not None:
            return Project(self.runt, node)

        node = await self.runt.getOneNode('proj:project', name, cmpr='^=')
        if node is not None:
            return Project(self.runt, node)

    async def _funcProjAdd(self, name, desc=''):

        gateiden = self.runt.snap.view.iden
        # do not use self.runt.confirm() to avoid asroot
        self.runt.user.confirm(('project', 'add'), gateiden=gateiden)

        tick = s_common.now()
        props = {
            'name': await tostr(name),
            'desc': await tostr(desc),
            'created': tick,
            'updated': tick,
            'creator': self.runt.user.iden,
        }
        node = await self.runt.snap.addNode('proj:project', '*', props=props)
        return Project(self.runt, node)



import copy
import regex
import argparse

import yaml

import synapse.exc as s_exc

import synapse.lib.cell as s_cell

import synapse.lib.config as s_config

import synapse.tests.utils as s_test

class SchemaCell(s_cell.Cell):
    confbase = {
        'apikey': {
            'description': 'fancy apikey',
            'type': 'string',
        },
        'apihost': {
            'description': 'host where the apikey goes too!',
            'type': 'string',
            'default': 'https://httpbin.org/'
        },
    }
    async def __anit__(self, dirn, conf=None, readonly=False, *args, **kwargs):
        await s_cell.Cell.__anit__(self, dirn, conf, readonly, *args, **kwargs)
        # This captures a design pattern that reduces boilerplate
        # code used by Cell implementators.
        self.conf.reqConfValu('apikey')


class ConfTest(s_test.SynTest):

    async def test_config_basics(self):

        conf = s_config.Config(s_test.test_schema)
        # Start out empty
        self.eq(conf.asDict(), {})

        # We can make an argparser that has config options populated in it
        # We explicitly skip boolean options without a default value so we
        # done end up in a ambiguous toggle case down the road.
        mesg = 'Boolean type is missing default information. ' \
               'Will not form argparse for [key:bool:nodefval]'
        pars = argparse.ArgumentParser('synapse.tests.test_lib_config.basics')
        pars.add_argument('--beep', type=str, help='beep option', default='beep.sys')
        with self.getLoggerStream('synapse.lib.config', mesg) as stream:
            for optname, optinfo in conf.getArgParseArgs():
                pars.add_argument(optname, **optinfo)
            self.true(stream.wait(3))
        hmsg = pars.format_help()

        # Undo pretty-printing
        hmsg = regex.sub(r'\s\s+', ' ', hmsg)

        # Multiple types are supported for argparse and descriptions
        # are used to generate the argparse help
        self.isin('--key-string KEY_STRING', hmsg)
        self.isin('Key String. I have a defval!', hmsg)
        self.isin('--key-integer KEY_INTEGER', hmsg)
        self.isin('--key-number KEY_NUMBER', hmsg)
        self.isin('--key-bool-defvalfalse', hmsg)
        self.isin('--key-bool-defvaltrue', hmsg)
        # The pre-existing parse is modified
        self.isin('--beep', hmsg)
        # We do not populate options for complex types
        self.notin('Key Array', hmsg)
        self.notin('Key Object', hmsg)
        # We do not populate options for bools with missing defaults
        self.notin('Key Bool, no default', hmsg)

        # And we can get the data too!  Unspecified values are set to
        # s_common.novalu so we know that they were NOT set at all.
        # This differs from the default argparse case of defaults being
        # set to None, and allows us to defer the injection of default
        # values to the jsonschema validation phase.
        args = ['--key-number', '1234.5678', '--key-bool-defvaltrue', 'false']
        opts = pars.parse_args(args)
        vopts = vars(opts)
        edata = {
            'key_bool_defvalfalse': None,
            'key_bool_defvaltrue': False,
            'key_integer': None,
            'key_number': 1234.5678,
            'key_string': None,
            'beep': 'beep.sys'
        }
        self.eq(edata, vopts)

        # We can re-inject the opts back into the config object.
        # The s_common.novalu data is skipped, as are opts which
        # were not set by the schema data.
        conf.setConfFromOpts(opts)
        self.eq(conf.asDict(), {
            'key:bool:defvaltrue': False,
            'key:number': 1234.5678,
        })

        # We can also get and load confdef data from environment
        # variables. These must be safe to decode as yaml. This
        # is mainly to faciliate machine base management of envars.
        a1 = yaml.safe_dump(['firetruck', 'spaceship'])
        i1 = yaml.safe_dump(8675309)
        n1 = yaml.safe_dump(9.813)
        # Load data from envars next - this shows precedence as well
        # where data already set won't be set again via this method.
        with self.setTstEnvars(KEY_ARRAY=a1,
                               KEY_NUMBER=n1,
                               KEY_INTEGER=i1,
                               ):
            conf.setConfFromEnvs()

        self.eq(conf.asDict(), {
            'key:bool:defvaltrue': False,
            'key:number': 1234.5678,
            'key:integer': 8675309,
            'key:array': ['firetruck', 'spaceship']
        })

        # we can set some remaining values directly
        conf.setdefault('key:object', {'rubber': 'ducky'})
        conf.setdefault('key:string', 'Funky string time!')
        self.eq(conf.asDict(), {
            'key:bool:defvaltrue': False,
            'key:number': 1234.5678,
            'key:integer': 8675309,
            'key:array': ['firetruck', 'spaceship'],
            'key:object': {'rubber': 'ducky'},
            'key:string': 'Funky string time!',
        })

        # Once we've built up our config, we can then ensure that it is valid.
        # This validation step also sets vars with defaults.  Keys without defaults
        # are not set at all.
        self.none(conf.reqConfValid())
        self.eq(conf.asDict(), {
            'key:bool:defvalfalse': False,
            'key:bool:defvaltrue': False,
            'key:number': 1234.5678,
            'key:integer': 8675309,
            'key:array': ['firetruck', 'spaceship'],
            'key:object': {'rubber': 'ducky'},
            'key:string': 'Funky string time!',
        })

        # We can ensure that certain vars are loaded
        self.eq('Funky string time!', conf.reqConfValu('key:string'))
        # And throw if they are not, or if the requested key isn't even schema valid
        self.raises(s_exc.NeedConfValu, conf.reqConfValu, 'key:bool:nodefval')
        self.raises(s_exc.BadArg, conf.reqConfValu, 'key:newp')

        # Since we're an Mutable mapping, we have some dict methods available to us
        self.len(7, conf)  # __len__
        self.eq(set(conf.keys()),  # __iter__
                {'key:bool:defvalfalse', 'key:bool:defvaltrue',
                 'key:number', 'key:integer', 'key:string',
                 'key:array', 'key:object',
                 })

        del conf['key:object']  # __delitem__
        self.eq(conf.asDict(), {
            'key:bool:defvalfalse': False,
            'key:bool:defvaltrue': False,
            'key:number': 1234.5678,
            'key:integer': 8675309,
            'key:array': ['firetruck', 'spaceship'],
            'key:string': 'Funky string time!',
        })

        # We have a convenience __repr__ :)
        valu = repr(conf)
        self.isin('<synapse.lib.config.Config at 0x', valu)
        self.isin('conf={', valu)

        # We can set invalid items and ensure the config is no longer valid.
        conf['key:array'] = 'Totally not an array.'
        self.raises(s_exc.BadConfValu, conf.reqConfValid)
        del conf['key:array']

        # We can do prefix-bassed collection of envar data.
        conf2 = s_config.Config(s_test.test_schema, envar_prefix='beeper')
        with self.setTstEnvars(BEEPER_KEY_ARRAY=a1,
                               KEY_INTEGER=i1,
                               ):
            conf2.setConfFromEnvs()
        # key:array is set, key:integer is not set.
        self.eq(conf2.asDict(), {
            'key:array': ['firetruck', 'spaceship']
        })

    async def test_config_fromcell(self):

        # We can make a conf from a cell ctor directly
        conf = s_config.Config.getConfFromCell(SchemaCell)
        self.isin('apikey', conf.json_schema.get('properties'))
        self.isin('apihost', conf.json_schema.get('properties'))

        # Assuming we populate that conf with some data
        # we can then use it to make a cell!

        conf['apikey'] = 'deadb33f'

        with self.getTestDir() as dirn:

            async with await SchemaCell.anit(dirn, conf=conf) as cell:
                self.eq(cell.conf.asDict(),
                        {'apikey': 'deadb33f',
                         'apihost': 'https://httpbin.org/'})

            # We can still make a cell with a dictionary being passed in directly.
            # The dictionary value is converted into the conf value.
            async with await SchemaCell.anit(dirn, conf={'apikey': 'deadb33f'}) as cell:
                self.eq(cell.conf.asDict(),
                        {'apikey': 'deadb33f',
                         'apihost': 'https://httpbin.org/'})

            with self.raises(s_exc.NeedConfValu) as cm:
                # Trying to make a cell with a missing key it wants fails
                async with await SchemaCell.anit(dirn, conf={}) as cell:
                    pass
            self.eq(cm.exception.get('key'), 'apikey')

    def test_hideconf(self):
        hide_schema = {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "additionalProperties": False,
            "properties": {
                'key:string': {
                    'description': 'Key String. I have a defval!',
                    'type': 'string',
                    'default': 'Default string!'
                },
                'key:integer': {
                    'description': 'Key Integer',
                    'type': 'integer',
                    'hideconf': True,
                },
            }
        }
        conf = s_config.Config(hide_schema)
        pars = argparse.ArgumentParser('synapse.tests.test_lib_config.test_hideconf')
        for optname, optinfo in conf.getArgParseArgs():
            pars.add_argument(optname, **optinfo)

        # key:integer is not present in cmdline
        hmsg = pars.format_help()
        self.isin('--key-string', hmsg)
        self.notin('--key-integer', hmsg)

        s1 = yaml.safe_dump('We all float down here')
        i1 = yaml.safe_dump(8675309)
        # Load data from envars next - this shows that we won't
        # set the key:integer value
        with self.setTstEnvars(KEY_STRING=s1,
                               KEY_INTEGER=i1,
                               ):
            conf.setConfFromEnvs()
        self.eq(conf.get('key:string'), 'We all float down here')
        self.none(conf.get('key:integer'))

        # hidearg instead of hideconf
        hide_schema = {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "additionalProperties": False,
            "properties": {
                'key:string': {
                    'description': 'Key String. I have a defval!',
                    'type': 'string',
                    'default': 'Default string!'
                },
                'key:integer': {
                    'description': 'Key Integer',
                    'type': 'integer',
                    'arg': True,
                    'hidecmdl': True,
                },
            }
        }
        conf = s_config.Config(hide_schema)
        pars = argparse.ArgumentParser('synapse.tests.test_lib_config.test_hideconf')
        for optname, optinfo in conf.getArgParseArgs():
            pars.add_argument(optname, **optinfo)

        # key:integer is not present in cmdline
        hmsg = pars.format_help()
        self.isin('--key-string', hmsg)
        self.notin('--key-integer', hmsg)

        s1 = yaml.safe_dump('We all float down here')
        i1 = yaml.safe_dump(8675309)
        # Load data from envars next - this shows we will
        # set the key:integer value
        with self.setTstEnvars(KEY_STRING=s1,
                               KEY_INTEGER=i1,
                               ):
            conf.setConfFromEnvs()
        self.eq(conf.get('key:string'), 'We all float down here')
        self.eq(conf.get('key:integer'), 8675309)

    def test_jsvalidator(self):

        schema = copy.deepcopy(s_test.test_schema)

        validator = s_config.getJsValidator(schema)
        item = validator({'key:number': 123})
        self.eq(item['key:number'], 123)
        self.eq(item['key:string'], 'Default string!')

        validator = s_config.getJsValidator(schema, use_default=False)
        item = validator({'key:number': 123})
        self.eq(item['key:number'], 123)
        self.notin('key:string', item)



import asyncio

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.tests.utils as s_test

import synapse.lib.hive as s_hive

tree0 = {
    'kids': {
        'hehe': {'value': 'haha'},
        'hoho': {'value': 'huhu', 'kids': {
            'foo': {'value': 99},
        }},
    }
}

tree1 = {
    'kids': {
        'hoho': {'value': 'huhu', 'kids': {
            'foo': {'value': 99},
        }}
    }
}

class HiveTest(s_test.SynTest):

    async def test_hive_slab(self):

        with self.getTestDir() as dirn:

            async with self.getTestHiveFromDirn(dirn) as hive:

                path = ('foo', 'bar')

                async with await hive.dict(path) as hivedict:

                    self.none(await hivedict.set('hehe', 200))
                    self.none(await hivedict.set('haha', 'hoho'))

                    valus = list(hivedict.values())
                    self.len(2, valus)
                    self.eq(set(valus), {200, 'hoho'})

                    self.eq(200, hivedict.get('hehe'))

                    self.eq(200, await hivedict.set('hehe', 300))

                    self.eq(300, hivedict.get('hehe'))

                    self.eq(300, await hive.get(('foo', 'bar', 'hehe')))
                    self.eq(300, await hive.set(('foo', 'bar', 'hehe'), 400))

                    hivedict.setdefault('lulz', 31337)

                    self.eq(31337, hivedict.get('lulz'))
                    await hivedict.set('lulz', 'boo')
                    items = list(hivedict.items())
                    self.eq([('hehe', 400), ('haha', 'hoho'), ('lulz', 'boo')], items)
                    self.eq('boo', await hivedict.pop('lulz'))
                    self.eq(31337, await hivedict.pop('lulz'))

                    self.eq(None, hivedict.get('nope'))

                    self.eq(s_common.novalu, hivedict.get('nope', default=s_common.novalu))
                    self.eq(s_common.novalu, await hivedict.pop('nope', default=s_common.novalu))

            async with self.getTestHiveFromDirn(dirn) as hive:

                self.eq(400, await hive.get(('foo', 'bar', 'hehe')))
                self.eq('hoho', await hive.get(('foo', 'bar', 'haha')))

                self.none(await hive.get(('foo', 'bar', 'lulz')))

    async def test_hive_telepath(self):

        # confirm that the primitives used by higher level APIs
        # work using telepath remotes and property synchronize.

        async with self.getTestHiveDmon() as dmon:

            turl = self.getTestUrl(dmon, 'hive')

            async with await s_hive.openurl(turl) as hive0:

                path = ('foo', 'bar')

                evnt = asyncio.Event()

                def onedit(mesg):
                    evnt.set()

                node0 = await hive0.open(path)
                node0.on('hive:set', onedit)

                async with await s_hive.openurl(turl) as hive1:

                    node1 = await hive1.open(path)
                    await node1.set(200)

                    await evnt.wait()

                    self.eq(200, node0.valu)

                    self.eq(201, await node0.add(1))
                    self.eq(202, await node1.add(1))
                    self.eq(203, await node0.add(1))

    async def test_hive_dir(self):

        async with self.getTestHive() as hive:

            await hive.open(('foo', 'bar'))
            await hive.open(('foo', 'baz'))
            await hive.open(('foo', 'faz'))

            self.none(hive.dir(('nosuchdir',)))

            self.eq([('foo', None, 3)], list(hive.dir(())))

            await hive.open(('foo',))

            kids = list(hive.dir(('foo',)))

            self.len(3, kids)

            names = list(sorted([name for (name, node, size) in kids]))

            self.eq(names, ('bar', 'baz', 'faz'))

    async def test_hive_pop(self):

        async with self.getTestHive() as hive:

            node = await hive.open(('foo', 'bar'))

            await node.set(20)

            self.none(await hive.pop(('newp',)))

            self.eq(20, await hive.pop(('foo', 'bar')))

            self.none(await hive.get(('foo', 'bar')))

            # Test recursive delete
            node = await hive.open(('foo', 'bar'))
            await node.set(20)

            self.eq(None, await hive.pop(('foo',)))
            self.none(await hive.get(('foo', 'bar')))

    async def test_hive_saveload(self):

        async with self.getTestHive() as hive:
            await hive.loadHiveTree(tree0)
            self.eq('haha', await hive.get(('hehe',)))
            self.eq('huhu', await hive.get(('hoho',)))
            self.eq(99, await hive.get(('hoho', 'foo')))

            await hive.loadHiveTree(tree1, trim=True)
            self.none(await hive.get(('hehe',)))
            self.eq('huhu', await hive.get(('hoho',)))
            self.eq(99, await hive.get(('hoho', 'foo')))

        async with self.getTestHive() as hive:

            node = await hive.open(('hehe', 'haha'))
            await node.set(99)

            tree = await hive.saveHiveTree()

            self.nn(tree['kids']['hehe'])
            self.nn(tree['kids']['hehe']['kids']['haha'])

            self.eq(99, tree['kids']['hehe']['kids']['haha']['value'])

    async def test_hive_exists(self):
        async with self.getTestHive() as hive:
            await hive.loadHiveTree(tree0)
            self.true(await hive.exists(('hoho', 'foo')))
            self.false(await hive.exists(('hoho', 'food')))
            self.false(await hive.exists(('newp',)))

    async def test_hive_rename(self):
        async with self.getTestHive() as hive:
            await hive.loadHiveTree(tree0)
            await self.asyncraises(s_exc.BadHivePath, hive.rename(('hehe',), ('hoho',)))
            await self.asyncraises(s_exc.BadHivePath, hive.rename(('newp',), ('newp2',)))
            await self.asyncraises(s_exc.BadHivePath, hive.rename(('hehe',), ('hehe', 'foo')))

            await hive.rename(('hehe',), ('lolo',))
            self.eq('haha', await hive.get(('lolo',)))
            self.false(await hive.exists(('hehe',)))

            await hive.rename(('hoho',), ('jojo',))
            self.false(await hive.exists(('hoho',)))
            jojo = await hive.open(('jojo',))
            self.len(1, jojo.kids)
            self.eq('huhu', jojo.valu)
            self.eq(99, await hive.get(('jojo', 'foo')))



import synapse.lib.datfile as s_datfile
import synapse.lib.msgpack as s_msgpack

def get(name, defval=None):
    '''
    Return an object from the embedded synapse data folder.

    Example:

        for tld in synapse.data.get('iana.tlds'):
            dostuff(tld)

    NOTE: Files are named synapse/data/<name>.mpk
    '''
    with s_datfile.openDatFile('synapse.data/%s.mpk' % name) as fd:
        return s_msgpack.un(fd.read())



import base64
import hashlib

import synapse.exc as s_exc
import synapse.common as s_common
import synapse.lib.types as s_types
import synapse.lib.module as s_module
import synapse.lookup.pe as s_l_pe
import synapse.lookup.macho as s_l_macho

class FileBase(s_types.Str):

    def postTypeInit(self):
        s_types.Str.postTypeInit(self)
        self.setNormFunc(str, self._normPyStr)

    def _normPyStr(self, valu):

        norm = valu.strip().lower().replace('\\', '/')
        if norm.find('/') != -1:
            mesg = 'file:base may not contain /'
            raise s_exc.BadTypeValu(name=self.name, valu=valu, mesg=mesg)

        subs = {}
        if norm.find('.') != -1:
            subs['ext'] = norm.rsplit('.', 1)[1]

        return norm, {'subs': subs}

class FilePath(s_types.Str):

    def postTypeInit(self):
        s_types.Str.postTypeInit(self)
        self.setNormFunc(str, self._normPyStr)

    def _normPyStr(self, valu):

        if len(valu) == 0:
            return '', {}

        lead = ''
        if valu[0] == '/':
            lead = '/'

        valu = valu.strip().lower().replace('\\', '/').strip('/')
        if not valu:
            return '', {}

        if valu in ('.', '..'):
            raise s_exc.BadTypeValu(name=self.name, valu=valu,
                                    mesg='Cannot norm a bare relative path.')

        path = []
        vals = [v for v in valu.split('/') if v]
        for part in vals:
            if part == '.':
                continue

            if part == '..':
                if len(path):
                    path.pop()

                continue

            path.append(part)

        fullpath = lead + '/'.join(path)

        base = path[-1]
        subs = {'base': base}
        if '.' in base:
            subs['ext'] = base.rsplit('.', 1)[1]
        if len(path) > 1:
            subs['dir'] = lead + '/'.join(path[:-1])

        return fullpath, {'subs': subs}

class FileBytes(s_types.Str):

    def postTypeInit(self):
        s_types.Str.postTypeInit(self)
        self.setNormFunc(str, self._normPyStr)
        self.setNormFunc(list, self._normPyList)
        self.setNormFunc(tuple, self._normPyList)
        self.setNormFunc(bytes, self._normPyBytes)

    def _normPyList(self, valu):
        guid, info = self.modl.type('guid').norm(valu)
        norm = f'guid:{guid}'
        return norm, {}

    def _normPyStr(self, valu):

        if valu == '*':
            guid = s_common.guid()
            norm = f'guid:{guid}'
            return norm, {}

        if valu.find(':') == -1:

            # we're ok with un-adorned sha256s
            if len(valu) == 64 and s_common.uhex(valu):
                valu = valu.lower()
                subs = {'sha256': valu}
                return f'sha256:{valu}', {'subs': subs}

            raise s_exc.BadTypeValu(name=self.name, valu=valu,
                                    mesg='unadorned file:bytes value is not a sha256')

        kind, kval = valu.split(':', 1)

        if kind == 'base64':
            byts = base64.b64decode(kval)
            return self._normPyBytes(byts)

        kval = kval.lower()

        if kind == 'hex':
            byts = s_common.uhex(kval)
            return self._normPyBytes(byts)

        if kind == 'guid':

            kval = kval.lower()
            if not s_common.isguid(kval):
                raise s_exc.BadTypeValu(name=self.name, valu=valu,
                                        mesg='guid is not a guid')

            return f'guid:{kval}', {}

        if kind == 'sha256':

            if len(kval) != 64:
                raise s_exc.BadTypeValu(name=self.name, valu=valu,
                                        mesg='invalid length for sha256 valu')

            s_common.uhex(kval)

            subs = {'sha256': kval}
            return f'sha256:{kval}', {'subs': subs}

        raise s_exc.BadTypeValu(name=self.name, valu=valu, kind=kind,
                                mesg='unable to norm as file:bytes')

    def _normPyBytes(self, valu):

        sha256 = hashlib.sha256(valu).hexdigest()

        norm = f'sha256:{sha256}'

        subs = {
            'md5': hashlib.md5(valu).hexdigest(),
            'sha1': hashlib.sha1(valu).hexdigest(),
            'sha256': sha256,
            'sha512': hashlib.sha512(valu).hexdigest(),
            'size': len(valu),
        }
        return norm, {'subs': subs}

class FileModule(s_module.CoreModule):

    async def initCoreModule(self):
        self.model.prop('file:bytes:mime').onSet(self._onSetFileBytesMime)
        self.core._setPropSetHook('file:bytes:sha256', self._hookFileBytesSha256)

    async def _hookFileBytesSha256(self, node, prop, norm):
        # this gets called post-norm and curv checks
        if node.ndef[1].startswith('sha256:'):
            mesg = "Can't change :sha256 on a file:bytes with sha256 based primary property."
            raise s_exc.BadTypeValu(mesg=mesg)

    async def _onSetFileBytesMime(self, node, oldv):
        name = node.get('mime')
        if name == '??':
            return
        await node.snap.addNode('file:ismime', (node.ndef[1], name))

    def getModelDefs(self):
        modl = {
            'ctors': (

                ('file:bytes', 'synapse.models.files.FileBytes', {}, {
                    'doc': 'The file bytes type with SHA256 based primary property.'}),

                ('file:base', 'synapse.models.files.FileBase', {}, {
                    'doc': 'A file name with no path.',
                    'ex': 'woot.exe'}),

                ('file:path', 'synapse.models.files.FilePath', {}, {
                    'doc': 'A normalized file path.',
                    'ex': 'c:/windows/system32/calc.exe'}),
            ),

            'interfaces': (
                ('file:mime:meta', {
                    'props': (
                        ('file', ('file:bytes', {}), {
                            'doc': 'The file that the mime info was parsed from.'}),
                        ('file:offs', ('int', {}), {
                            'doc': 'The optional offset where the mime info was parsed from.'}),
                        ('file:data', ('data', {}), {
                            'doc': 'A mime specific arbitrary data structure for non-indexed data.',
                        }),
                    ),
                    'doc': 'Properties common to mime specific file metadata types.',
                }),
                ('file:mime:msoffice', {
                    'props': (
                        ('title', ('str', {}), {
                            'doc': 'The title extracted from Microsoft Office metadata.'}),
                        ('author', ('str', {}), {
                            'doc': 'The author extracted from Microsoft Office metadata.'}),
                        ('subject', ('str', {}), {
                            'doc': 'The subject extracted from Microsoft Office metadata.'}),
                        ('application', ('str', {}), {
                            'doc': 'The creating_application extracted from Microsoft Office metadata.'}),
                        ('created', ('time', {}), {
                            'doc': 'The create_time extracted from Microsoft Office metadata.'}),
                        ('lastsaved', ('time', {}), {
                            'doc': 'The last_saved_time extracted from Microsoft Office metadata.'}),
                    ),
                    'doc': 'Properties common to various microsoft office file formats.',
                    'interfaces': ('file:mime:meta',),
                }),
                ('file:mime:image', {
                    'props': (
                        ('desc', ('str', {}), {
                            'doc': 'MIME specific description field extracted from metadata.'}),
                        ('comment', ('str', {}), {
                            'doc': 'MIME specific comment field extracted from metadata.'}),
                        ('created', ('time', {}), {
                            'doc': 'MIME specific creation timestamp extracted from metadata.'}),
                        ('imageid', ('str', {}), {
                            'doc': 'MIME specific unique identifier extracted from metadata.'}),
                        ('author', ('ps:contact', {}), {
                            'doc': 'MIME specific contact information extracted from metadata.'}),
                        ('latlong', ('geo:latlong', {}), {
                            'doc': 'MIME specific lat/long information extracted from metadata.'}),
                        ('altitude', ('geo:altitude', {}), {
                            'doc': 'MIME specific altitude information extracted from metadata.'}),
                    ),
                    'doc': 'Properties common to image file formats.',
                    'interfaces': ('file:mime:meta',),
                }),
                ('file:mime:macho:loadcmd', {
                    'props': (
                        ('file', ('file:bytes', {}), {
                            'doc': 'The Mach-O file containing the load command.'}),
                        ('type', ('int', {'enums': s_l_macho.getLoadCmdTypes()}), {
                            'doc': 'The type of the load command'}),
                        ('size', ('int', {}), {
                            'doc': 'The size of the load command structure in bytes.'}),
                    ),
                    'doc': 'Properties common to all Mach-O load commands',
                })
            ),

            'types': (

                ('file:subfile', ('comp', {'fields': (('parent', 'file:bytes'), ('child', 'file:bytes'))}), {
                    'doc': 'A parent file that fully contains the specified child file.',
                }),

                ('file:filepath', ('comp', {'fields': (('file', 'file:bytes'), ('path', 'file:path'))}), {
                    'doc': 'The fused knowledge of the association of a file:bytes node and a file:path.',
                }),

                ('file:mime', ('str', {'lower': 1}), {
                    'doc': 'A file mime name string.',
                    'ex': 'text/plain',
                }),

                ('file:ismime', ('comp', {'fields': (('file', 'file:bytes'), ('mime', 'file:mime'))}), {
                    'doc': 'Records one, of potentially multiple, mime types for a given file.',
                }),

                ('file:mime:msdoc', ('guid', {}), {
                    'doc': 'The GUID of a set of mime metadata for a Microsoft Word file.',
                    'interfaces': ('file:mime:msoffice',),
                }),

                ('file:mime:msxls', ('guid', {}), {
                    'doc': 'The GUID of a set of mime metadata for a Microsoft Excel file.',
                    'interfaces': ('file:mime:msoffice',),
                }),

                ('file:mime:msppt', ('guid', {}), {
                    'doc': 'The GUID of a set of mime metadata for a Microsoft Powerpoint file.',
                    'interfaces': ('file:mime:msoffice',),
                }),

                ('file:mime:rtf', ('guid', {}), {
                    'doc': 'The GUID of a set of mime metadata for a .rtf file.',
                    'interfaces': ('file:mime:meta',),
                }),

                ('file:mime:jpg', ('guid', {}), {
                    'doc': 'The GUID of a set of mime metadata for a .jpg file.',
                    'interfaces': ('file:mime:image',),
                }),

                ('file:mime:tif', ('guid', {}), {
                    'doc': 'The GUID of a set of mime metadata for a .tif file.',
                    'interfaces': ('file:mime:image',),
                }),

                ('file:mime:gif', ('guid', {}), {
                    'doc': 'The GUID of a set of mime metadata for a .gif file.',
                    'interfaces': ('file:mime:image',),
                }),

                ('file:mime:png', ('guid', {}), {
                    'doc': 'The GUID of a set of mime metadata for a .png file.',
                    'interfaces': ('file:mime:image',),
                }),

                ('file:mime:pe:section', ('comp', {'fields': (
                        ('file', 'file:bytes'),
                        ('name', 'str'),
                        ('sha256', 'hash:sha256'),
                    )}), {
                    'doc': 'The fused knowledge a file:bytes node containing a pe section.',
                }),
                ('file:mime:pe:resource', ('comp', {'fields': (
                        ('file', 'file:bytes'),
                        ('type', 'pe:resource:type'),
                        ('langid', 'pe:langid'),
                        ('resource', 'file:bytes'))}), {
                    'doc': 'The fused knowledge of a file:bytes node containing a pe resource.',
                }),
                ('file:mime:pe:export', ('comp', {'fields': (
                        ('file', 'file:bytes'),
                        ('name', 'str'))}), {
                    'doc': 'The fused knowledge of a file:bytes node containing a pe named export.',
                }),
                ('file:mime:pe:vsvers:keyval', ('comp', {'fields': (
                        ('name', 'str'),
                        ('value', 'str'))}), {
                    'doc': 'A key value pair found in a PE vsversion info structure.',
                }),
                ('file:mime:pe:vsvers:info', ('comp', {'fields': (
                        ('file', 'file:bytes'),
                        ('keyval', 'file:mime:pe:vsvers:keyval'))}), {
                    'doc': 'knowledge of a file:bytes node containing vsvers info.',
                }),
                ('file:string', ('comp', {'fields': (
                        ('file', 'file:bytes'),
                        ('string', 'str'))}), {
                    'deprecated': True,
                    'doc': 'Deprecated. Please use the edge -(refs)> it:dev:str.',
                }),

                ('pe:resource:type', ('int', {'enums': s_l_pe.getRsrcTypes()}), {
                    'doc': 'The typecode for the resource.',
                }),

                ('pe:langid', ('int', {'enums': s_l_pe.getLangCodes()}), {
                    'doc': 'The PE language id.',
                }),

                ('file:mime:macho:loadcmd', ('guid', {}), {
                    'doc': 'A generic load command pulled from the Mach-O headers.',
                    'interfaces': ('file:mime:macho:loadcmd',),
                }),

                ('file:mime:macho:version', ('guid', {}), {
                    'doc': 'A specific load command used to denote the version of the source used to build the Mach-O binary.',
                    'interfaces': ('file:mime:macho:loadcmd',),
                }),

                ('file:mime:macho:uuid', ('guid', {}), {
                    'doc': 'A specific load command denoting a UUID used to uniquely identify the Mach-O binary.',
                    'interfaces': ('file:mime:macho:loadcmd',),
                }),

                ('file:mime:macho:segment', ('guid', {}), {
                    'doc': 'A named region of bytes inside a Mach-O binary.',
                    'interfaces': ('file:mime:macho:loadcmd',),
                }),

                ('file:mime:macho:section', ('guid', {}), {
                    'doc': 'A section inside a Mach-O binary denoting a named region of bytes inside a segment.',
                }),

            ),

            'forms': (

                ('file:bytes', {}, (

                    ('size', ('int', {}), {'doc': 'The file size in bytes.'}),

                    ('md5', ('hash:md5', {}), {'doc': 'The md5 hash of the file.'}),

                    ('sha1', ('hash:sha1', {}), {'doc': 'The sha1 hash of the file.'}),

                    ('sha256', ('hash:sha256', {}), {'doc': 'The sha256 hash of the file.'}),

                    ('sha512', ('hash:sha512', {}), {'doc': 'The sha512 hash of the file.'}),

                    ('name', ('file:base', {}), {
                        'doc': 'The best known base name for the file.'}),

                    ('mime', ('file:mime', {}), {
                        'doc': 'The "best" mime type name for the file.'}),

                    ('mime:x509:cn', ('str', {}), {
                        'doc': 'The Common Name (CN) attribute of the x509 Subject.'}),

                    ('mime:pe:size', ('int', {}), {
                        'doc': 'The size of the executable file according to the PE file header.'}),

                    ('mime:pe:imphash', ('guid', {}), {
                        'doc': 'The PE import hash of the file as calculated by pefile; '
                               'https://github.com/erocarrera/pefile .'}),

                    ('mime:pe:compiled', ('time', {}), {
                        'doc': 'The compile time of the file according to the PE header.'}),

                    ('mime:pe:pdbpath', ('file:path', {}), {
                        'doc': 'The PDB string according to the PE.'}),

                    ('mime:pe:exports:time', ('time', {}), {
                        'doc': 'The export time of the file according to the PE.'}),

                    ('mime:pe:exports:libname', ('str', {}), {
                        'doc': 'The export library name according to the PE.'}),

                    ('mime:pe:richhdr', ('hash:sha256', {}), {
                        'doc': 'The sha256 hash of the rich header bytes.'}),

                )),

                ('file:mime', {}, ()),

                ('file:ismime', {}, (
                    ('file', ('file:bytes', {}), {
                        'ro': True,
                        'doc': 'The file node that is an instance of the named mime type.',
                    }),
                    ('mime', ('file:mime', {}), {
                        'ro': True,
                        'doc': 'The mime type of the file.',
                    }),
                )),

                ('file:mime:msdoc', {}, ()),
                ('file:mime:msxls', {}, ()),
                ('file:mime:msppt', {}, ()),

                ('file:mime:jpg', {}, ()),
                ('file:mime:tif', {}, ()),
                ('file:mime:gif', {}, ()),
                ('file:mime:png', {}, ()),

                ('file:mime:rtf', {}, (
                    ('guid', ('guid', {}), {
                        'doc': 'The parsed GUID embedded in the .rtf file.'}),
                )),

                ('file:mime:pe:section', {}, (
                    ('file', ('file:bytes', {}), {
                        'ro': True,
                        'doc': 'The file containing the section.',
                    }),
                    ('name', ('str', {}), {
                        'ro': True,
                        'doc': 'The textual name of the section.',
                    }),
                    ('sha256', ('hash:sha256', {}), {
                        'ro': True,
                        'doc': 'The sha256 hash of the section. Relocations must be zeroed before hashing.',
                    }),
                )),

                ('file:mime:pe:resource', {}, (
                    ('file', ('file:bytes', {}), {
                        'ro': True,
                        'doc': 'The file containing the resource.',
                    }),
                    ('type', ('pe:resource:type', {}), {
                        'ro': True,
                        'doc': 'The typecode for the resource.',
                    }),
                    ('langid', ('pe:langid', {}), {
                        'ro': True,
                        'doc': 'The language code for the resource.',
                    }),
                    ('resource', ('file:bytes', {}), {
                        'ro': True,
                        'doc': 'The sha256 hash of the resource bytes.',
                    }),
                )),

                ('file:mime:pe:export', {}, (
                    ('file', ('file:bytes', {}), {
                        'ro': True,
                        'doc': 'The file containing the export.',
                    }),
                    ('name', ('str', {}), {
                        'ro': True,
                        'doc': 'The name of the export in the file.',
                    }),
                )),

                ('file:mime:pe:vsvers:keyval', {}, (
                    ('name', ('str', {}), {
                        'ro': True,
                        'doc': 'The key for the vsversion keyval pair.',
                    }),
                    ('value', ('str', {}), {
                        'ro': True,
                        'doc': 'The value for the vsversion keyval pair.',
                    }),
                )),

                ('file:mime:pe:vsvers:info', {}, (
                    ('file', ('file:bytes', {}), {
                        'ro': True,
                        'doc': 'The file containing the vsversion keyval pair.',
                    }),
                    ('keyval', ('file:mime:pe:vsvers:keyval', {}), {
                        'ro': True,
                        'doc': 'The vsversion info keyval in this file:bytes node.',
                    }),
                )),

                ('file:string', {}, (
                    ('file', ('file:bytes', {}), {
                        'ro': True,
                        'doc': 'The file containing the string.',
                    }),
                    ('string', ('str', {}), {
                        'ro': True,
                        'doc': 'The string contained in this file:bytes node.',
                    }),
                )),

                ('file:base', {}, (
                    ('ext', ('str', {}), {'ro': True,
                        'doc': 'The file extension (if any).'}),
                )),

                ('file:filepath', {}, (
                    ('file', ('file:bytes', {}), {
                        'ro': True,
                        'doc': 'The file seen at a path.',
                    }),
                    ('path', ('file:path', {}), {
                        'ro': True,
                        'doc': 'The path a file was seen at.',
                    }),
                    ('path:dir', ('file:path', {}), {
                        'ro': True,
                        'doc': 'The parent directory.',
                    }),
                    ('path:base', ('file:base', {}), {
                        'ro': True,
                        'doc': 'The name of the file.',
                    }),
                    ('path:base:ext', ('str', {}), {
                        'ro': True,
                        'doc': 'The extension of the file name.',
                    }),
                )),

                ('file:subfile', {}, (
                    ('parent', ('file:bytes', {}), {
                        'ro': True,
                        'doc': 'The parent file containing the child file.',
                    }),
                    ('child', ('file:bytes', {}), {
                        'ro': True,
                        'doc': 'The child file contained in the parent file.',
                    }),
                    ('name', ('file:base', {}), {
                        'deprecated': True,
                        'doc': 'Deprecated, please use the :path property.',
                    }),
                    ('path', ('file:path', {}), {
                        'doc': 'The path that the parent uses to refer to the child file.',
                    }),
                )),

                ('file:path', {}, (
                    ('dir', ('file:path', {}), {'ro': True,
                        'doc': 'The parent directory.'}),

                    ('base', ('file:base', {}), {'ro': True,
                        'doc': 'The file base name.'}),

                    ('base:ext', ('str', {}), {'ro': True,
                        'doc': 'The file extension.'}),
                )),

                ('file:mime:macho:loadcmd', {}, ()),
                ('file:mime:macho:version', {}, (
                    ('version', ('str', {}), {
                        'doc': 'The version of the Mach-O file encoded in an LC_VERSION load command.'}),
                )),
                ('file:mime:macho:uuid', {}, (
                    ('uuid', ('guid', {}), {
                        'doc': 'The UUID of the Mach-O application (as defined in an LC_UUID load command).'}),
                )),
                ('file:mime:macho:segment', {}, (
                    ('name', ('str', {}), {
                        'doc': 'The name of the Mach-O segment.'}),
                    ('memsize', ('int', {}), {
                        'doc': 'The size of the segment in bytes, when resident in memory, according to the load command structure.'}),
                    ('disksize', ('int', {}), {
                        'doc': 'The size of the segment in bytes, when on disk, according to the load command structure.'}),
                    ('sha256', ('hash:sha256', {}), {
                        'doc': 'The sha256 hash of the bytes of the segment.'}),
                    ('offset', ('int', {}), {
                        'doc': 'The file offset to the begining of the segment.'}),
                )),
                ('file:mime:macho:section', {}, (
                    ('segment', ('file:mime:macho:segment', {}), {
                        'doc': 'The Mach-O segment that contains this section.'}),
                    ('name', ('str', {}), {
                        'doc': 'Name of the section.'}),
                    ('size', ('int', {}), {
                        'doc': 'Size of the section in bytes.'}),
                    ('type', ('int', {'enums': s_l_macho.getSectionTypes()}), {
                        'doc': 'The type of the section.'}),
                    ('sha256', ('hash:sha256', {}), {
                        'doc': 'The sha256 hash of the bytes of the Mach-O section.'}),
                    ('offset', ('int', {}), {
                        'doc': 'The file offset to the begining of the section'}),
                )),

            ),

        }

        name = 'file'
        return ((name, modl),)



import synapse.exc as s_exc
import synapse.common as s_common

import synapse.models.crypto as s_m_crypto

import synapse.tests.utils as s_t_utils

class InfotechModelTest(s_t_utils.SynTest):

    async def test_infotech_basics(self):

        async with self.getTestCore() as core:

            nodes = await core.nodes('''[
                it:sec:cwe=CWE-120
                    :name=omg
                    :desc=omgwtfbbq
                    :url=https://cwe.mitre.org/data/definitions/120.html
                    :parents=(CWE-119,)
            ]''')
            self.len(1, nodes)
            self.eq(nodes[0].ndef, ('it:sec:cwe', 'CWE-120'))
            self.eq(nodes[0].get('name'), 'omg')
            self.eq(nodes[0].get('desc'), 'omgwtfbbq')
            self.eq(nodes[0].get('url'), 'https://cwe.mitre.org/data/definitions/120.html')
            self.eq(nodes[0].get('parents'), ('CWE-119',))

            self.eq(r'foo\:bar', core.model.type('it:sec:cpe').norm(r'cpe:2.3:a:foo\:bar:*:*:*:*:*:*:*:*:*')[1]['subs']['vendor'])

            with self.raises(s_exc.BadTypeValu):
                nodes = await core.nodes('[it:sec:cpe=asdf]')

            with self.raises(s_exc.BadTypeValu):
                nodes = await core.nodes('[it:sec:cpe=cpe:2.3:a:asdf]')

            nodes = await core.nodes('''[
                it:sec:cpe=cpe:2.3:a:microsoft:internet_explorer:8.0.6001:beta:*:*:*:*:*:*
            ]''')
            self.len(1, nodes)
            self.eq(nodes[0].ndef, ('it:sec:cpe', 'cpe:2.3:a:microsoft:internet_explorer:8.0.6001:beta:*:*:*:*:*:*'))
            self.eq(nodes[0].get('part'), 'a')
            self.eq(nodes[0].get('vendor'), 'microsoft')
            self.eq(nodes[0].get('product'), 'internet_explorer')
            self.eq(nodes[0].get('version'), '8.0.6001')
            self.eq(nodes[0].get('update'), 'beta')

            nodes = await core.nodes('''[
                it:mitre:attack:group=G0100
                    :org={[ ou:org=* :name=visicorp ]}
                    :name=aptvisi
                    :names=(visigroup, nerdsrus, visigroup)
                    :desc=worlddom
                    :url=https://vertex.link
                    :tag=cno.mitre.g0100
                    :references=(https://foo.com,https://bar.com)
                    :software=(S0200,S0100,S0100)
                    :techniques=(T0200,T0100,T0100)
            ]''')
            self.len(1, nodes)
            self.eq(nodes[0].ndef, ('it:mitre:attack:group', 'G0100'))
            self.nn(nodes[0].get('org'))
            self.eq(nodes[0].get('name'), 'aptvisi')
            self.eq(nodes[0].get('names'), ('nerdsrus', 'visigroup'))
            self.eq(nodes[0].get('desc'), 'worlddom')
            self.eq(nodes[0].get('tag'), 'cno.mitre.g0100')
            self.eq(nodes[0].get('url'), 'https://vertex.link')
            self.eq(nodes[0].get('references'), ('https://foo.com', 'https://bar.com'))
            self.eq(nodes[0].get('software'), ('S0100', 'S0200'))
            self.eq(nodes[0].get('techniques'), ('T0100', 'T0200'))

            nodes = await core.nodes('''[
                it:mitre:attack:tactic=TA0100
                    :name=tactilneck
                    :desc=darkerblack
                    :url=https://archer.link
                    :tag=cno.mitre.ta0100
                    :references=(https://foo.com,https://bar.com)
            ]''')
            self.len(1, nodes)
            self.eq(nodes[0].ndef, ('it:mitre:attack:tactic', 'TA0100'))
            self.eq(nodes[0].get('name'), 'tactilneck')
            self.eq(nodes[0].get('desc'), 'darkerblack')
            self.eq(nodes[0].get('tag'), 'cno.mitre.ta0100')
            self.eq(nodes[0].get('url'), 'https://archer.link')
            self.eq(nodes[0].get('references'), ('https://foo.com', 'https://bar.com'))

            nodes = await core.nodes('''[
                it:mitre:attack:technique=T0100
                    :name=lockpicking
                    :desc=speedhackers
                    :url=https://locksrus.link
                    :tag=cno.mitre.t0100
                    :references=(https://foo.com,https://bar.com)
                    :parent=T9999
                    :status=deprecated
                    :isnow=T1110
                    :tactics=(TA0200,TA0100,TA0100)
            ]''')
            self.len(1, nodes)
            self.eq(nodes[0].ndef, ('it:mitre:attack:technique', 'T0100'))
            self.eq(nodes[0].get('name'), 'lockpicking')
            self.eq(nodes[0].get('desc'), 'speedhackers')
            self.eq(nodes[0].get('tag'), 'cno.mitre.t0100')
            self.eq(nodes[0].get('url'), 'https://locksrus.link')
            self.eq(nodes[0].get('references'), ('https://foo.com', 'https://bar.com'))
            self.eq(nodes[0].get('parent'), 'T9999')
            self.eq(nodes[0].get('tactics'), ('TA0100', 'TA0200'))
            self.eq(nodes[0].get('status'), 'deprecated')
            self.eq(nodes[0].get('isnow'), 'T1110')

            nodes = await core.nodes('''[
                it:mitre:attack:software=S0100
                    :software=*
                    :name=redtree
                    :names=("redtree alt", eviltree)
                    :desc=redtreestuff
                    :url=https://redtree.link
                    :tag=cno.mitre.s0100
                    :references=(https://foo.com,https://bar.com)
                    :techniques=(T0200,T0100,T0100)
            ]''')
            self.len(1, nodes)
            self.eq(nodes[0].ndef, ('it:mitre:attack:software', 'S0100'))
            self.nn(nodes[0].get('software'))
            self.eq(nodes[0].get('name'), 'redtree')
            self.eq(nodes[0].get('names'), ('eviltree', 'redtree alt'))
            self.eq(nodes[0].get('desc'), 'redtreestuff')
            self.eq(nodes[0].get('tag'), 'cno.mitre.s0100')
            self.eq(nodes[0].get('url'), 'https://redtree.link')
            self.eq(nodes[0].get('references'), ('https://foo.com', 'https://bar.com'))
            self.eq(nodes[0].get('techniques'), ('T0100', 'T0200'))

            nodes = await core.nodes('''[
                it:mitre:attack:mitigation=M0100
                    :name=patchstuff
                    :desc=patchyourstuff
                    :url=https://wsus.com
                    :tag=cno.mitre.m0100
                    :references=(https://foo.com,https://bar.com)
                    :addresses=(T0200,T0100,T0100)
            ]''')
            self.len(1, nodes)
            self.eq(nodes[0].ndef, ('it:mitre:attack:mitigation', 'M0100'))
            self.eq(nodes[0].get('name'), 'patchstuff')
            self.eq(nodes[0].get('desc'), 'patchyourstuff')
            self.eq(nodes[0].get('tag'), 'cno.mitre.m0100')
            self.eq(nodes[0].get('url'), 'https://wsus.com')
            self.eq(nodes[0].get('references'), ('https://foo.com', 'https://bar.com'))
            self.eq(nodes[0].get('addresses'), ('T0100', 'T0200'))

            nodes = await core.nodes('''[
                it:exec:thread=*
                    :proc=*
                    :created=20210202
                    :exited=20210203
                    :exitcode=0
                    :src:proc=*
                    :src:thread=*
            ]''')
            self.len(1, nodes)
            self.nn(nodes[0].ndef[1])
            self.eq(nodes[0].get('created'), 1612224000000)
            self.eq(nodes[0].get('exited'), 1612310400000)
            self.eq(nodes[0].get('exitcode'), 0)
            self.len(1, await core.nodes('it:exec:thread:created :proc -> it:exec:proc'))
            self.len(1, await core.nodes('it:exec:thread:created :src:proc -> it:exec:proc'))
            self.len(1, await core.nodes('it:exec:thread:created :src:thread -> it:exec:thread'))

            nodes = await core.nodes('''[
                it:exec:loadlib=*
                    :proc=*
                    :va=0x00a000
                    :loaded=20210202
                    :unloaded=20210203
                    :path=/home/invisigoth/rootkit.so
                    :file=*
            ]''')
            self.len(1, nodes)
            self.nn(nodes[0].ndef[1])
            self.nn(nodes[0].get('proc'))
            self.eq(nodes[0].get('va'), 0x00a000)
            self.eq(nodes[0].get('loaded'), 1612224000000)
            self.eq(nodes[0].get('unloaded'), 1612310400000)
            self.len(1, await core.nodes('it:exec:loadlib -> file:bytes'))
            self.len(1, await core.nodes('it:exec:loadlib :proc -> it:exec:proc'))
            self.len(1, await core.nodes('it:exec:loadlib -> file:path +file:path=/home/invisigoth/rootkit.so'))

            nodes = await core.nodes('''[
                it:exec:mmap=*
                    :proc=*
                    :va=0x00a000
                    :size=4096
                    :perms:read=1
                    :perms:write=0
                    :perms:execute=1
                    :created=20210202
                    :deleted=20210203
                    :path=/home/invisigoth/rootkit.so
                    :hash:sha256=ad9f4fe922b61e674a09530831759843b1880381de686a43460a76864ca0340c
            ]''')
            self.len(1, nodes)
            self.nn(nodes[0].ndef[1])
            self.nn(nodes[0].get('proc'))
            self.eq(nodes[0].get('va'), 0x00a000)
            self.eq(nodes[0].get('size'), 4096)
            self.eq(nodes[0].get('perms:read'), 1)
            self.eq(nodes[0].get('perms:write'), 0)
            self.eq(nodes[0].get('perms:execute'), 1)
            self.eq(nodes[0].get('created'), 1612224000000)
            self.eq(nodes[0].get('deleted'), 1612310400000)
            self.eq(nodes[0].get('hash:sha256'), 'ad9f4fe922b61e674a09530831759843b1880381de686a43460a76864ca0340c')
            self.len(1, await core.nodes('it:exec:mmap -> hash:sha256'))
            self.len(1, await core.nodes('it:exec:mmap :proc -> it:exec:proc'))
            self.len(1, await core.nodes('it:exec:mmap -> file:path +file:path=/home/invisigoth/rootkit.so'))

            nodes = await core.nodes('''[
                it:exec:proc=80e6c59d9c349ac15f716eaa825a23fa
                    :killedby=*
                    :exitcode=0
                    :exited=20210202
            ]''')
            self.len(1, nodes)
            self.eq(nodes[0].ndef[1], '80e6c59d9c349ac15f716eaa825a23fa')
            self.nn(nodes[0].get('killedby'))
            self.eq(nodes[0].get('exitcode'), 0)
            self.eq(nodes[0].get('exited'), 1612224000000)
            self.len(1, await core.nodes('it:exec:proc=80e6c59d9c349ac15f716eaa825a23fa :killedby -> it:exec:proc'))

            nodes = await core.nodes('''[
                it:av:prochit=*
                    :proc=*
                    :sig=(a6834cea191af070abb11af59d881c40, 'foobar')
                    :time=20210202
            ]''')
            self.len(1, nodes)
            self.nn(nodes[0].ndef[1])
            self.nn(nodes[0].get('proc'))
            self.eq(nodes[0].get('sig'), ('a6834cea191af070abb11af59d881c40', 'foobar'))
            self.eq(nodes[0].get('time'), 1612224000000)
            self.len(1, await core.nodes('it:av:prochit -> it:av:sig'))
            self.len(1, await core.nodes('it:av:prochit -> it:exec:proc'))

            nodes = await core.nodes('''[
                it:app:yara:procmatch=*
                    :proc=*
                    :rule=*
                    :time=20210202
            ]''')
            self.len(1, nodes)
            self.nn(nodes[0].ndef[1])
            self.nn(nodes[0].get('proc'))
            self.nn(nodes[0].get('rule'))
            self.eq(nodes[0].get('time'), 1612224000000)
            self.len(1, await core.nodes('it:app:yara:procmatch -> it:exec:proc'))
            self.len(1, await core.nodes('it:app:yara:procmatch -> it:app:yara:rule'))

    async def test_infotech_ios(self):

        async with self.getTestCore() as core:

            async with await core.snap() as snap:

                valu = '00000000-0000-0000-0000-00000000000A'
                idfa = await snap.addNode('it:os:ios:idfa', valu)
                self.eq(idfa.ndef[1], '00000000-0000-0000-0000-00000000000a')

    async def test_infotech_android(self):

        softver = s_common.guid()

        async with self.getTestCore() as core:

            async with await core.snap() as snap:

                perm = await snap.addNode('it:os:android:perm', 'Foo Perm')
                self.eq(perm.ndef[1], 'Foo Perm')
                intent = await snap.addNode('it:os:android:intent', 'Foo Intent')
                self.eq(intent.ndef[1], 'Foo Intent')

                ilisn = await snap.addNode('it:os:android:ilisten', (softver, 'Listen Test'))
                self.eq(ilisn.get('app'), softver)
                self.eq(ilisn.get('intent'), 'Listen Test')

                ibcast = await snap.addNode('it:os:android:ibroadcast', (softver, 'Broadcast Test'))
                self.eq(ibcast.get('app'), softver)
                self.eq(ibcast.get('intent'), 'Broadcast Test')

                reqperm = await snap.addNode('it:os:android:reqperm', (softver, 'Test Perm'))
                self.eq(reqperm.get('app'), softver)
                self.eq(reqperm.get('perm'), 'Test Perm')

                valu = 'someIdentifier'
                aaid = await snap.addNode('it:os:android:aaid', valu)
                self.eq(aaid.ndef[1], 'someidentifier')

    async def test_it_forms_simple(self):
        async with self.getTestCore() as core:
            place = s_common.guid()
            async with await core.snap() as snap:
                node = await snap.addNode('it:hostname', 'Bobs Computer')
                self.eq(node.ndef[1], 'bobs computer')
                org0 = s_common.guid()
                host0 = s_common.guid()
                sver0 = s_common.guid()
                cont0 = s_common.guid()
                hprops = {
                    'name': 'Bobs laptop',
                    'desc': 'Bobs paperweight',
                    'ipv4': '1.2.3.4',
                    'latlong': '0.0, 0.0',
                    'place': place,
                    'os': sver0,
                    'manu': 'Dull',
                    'model': 'Lutitude 8249',
                    'serial': '111-222',
                    'loc': 'us.hehe.haha',
                    'operator': cont0,
                    'org': org0,
                }
                node = await snap.addNode('it:host', host0, hprops)
                self.eq(node.ndef[1], host0)
                self.eq(node.get('name'), 'bobs laptop')
                self.eq(node.get('desc'), 'Bobs paperweight')
                self.eq(node.get('ipv4'), 0x01020304)
                self.eq(node.get('latlong'), (0.0, 0.0))
                self.eq(node.get('place'), place)
                self.eq(node.get('os'), sver0)
                self.eq(node.get('loc'), 'us.hehe.haha')
                self.eq(node.get('org'), org0)
                self.eq(node.get('operator'), cont0)

                node = await snap.addNode('it:hosturl', (host0, 'http://vertex.ninja/cool.php'))
                self.eq(node.ndef[1], (host0, 'http://vertex.ninja/cool.php'))
                self.eq(node.get('host'), host0)
                self.eq(node.get('url'), 'http://vertex.ninja/cool.php')

                node = await snap.addNode('it:dev:int', 0x61C88648)
                self.eq(node.ndef[1], 1640531528)

                cprops = {
                    'desc': 'Some words.',
                }
                node = await snap.addNode('it:sec:cve', 'CVE-2013-9999', cprops)
                self.eq(node.ndef[1], 'cve-2013-9999')
                self.eq(node.get('desc'), 'Some words.')

                hash0 = s_common.guid()
                hprops = {
                    'salt': 'B33F',
                    'hash:md5': s_m_crypto.ex_md5,
                    'hash:sha1': s_m_crypto.ex_sha1,
                    'hash:sha256': s_m_crypto.ex_sha256,
                    'hash:sha512': s_m_crypto.ex_sha512,
                    'hash:lm': s_m_crypto.ex_md5,
                    'hash:ntlm': s_m_crypto.ex_md5,
                    'passwd': "I've got the same combination on my luggage!",
                }
                node = await snap.addNode('it:auth:passwdhash', hash0, hprops)
                self.eq(node.ndef[1], hash0)
                self.eq(node.get('salt'), 'b33f')
                self.eq(node.get('hash:md5'), s_m_crypto.ex_md5)
                self.eq(node.get('hash:sha1'), s_m_crypto.ex_sha1)
                self.eq(node.get('hash:sha256'), s_m_crypto.ex_sha256)
                self.eq(node.get('hash:sha512'), s_m_crypto.ex_sha512)
                self.eq(node.get('hash:lm'), s_m_crypto.ex_md5)
                self.eq(node.get('hash:ntlm'), s_m_crypto.ex_md5)
                self.eq(node.get('passwd'), "I've got the same combination on my luggage!")

            nodes = await core.nodes('[ it:adid=visi ]')
            self.eq(('it:adid', 'visi'), nodes[0].ndef)

            nodes = await core.nodes('''
                init {
                    $org = $lib.guid()
                    $host = $lib.guid()
                    $acct = $lib.guid()
                }
                [
                    it:account=$acct
                        :host=$host
                        :user=visi
                        :contact={[ ps:contact=* :email=visi@vertex.link ]}
                        :domain={[ it:domain=* :org=$org :name=vertex :desc="the vertex project domain" ]}

                    (it:logon=* :time=20210314 :logoff:time=202103140201 :account=$acct :host=$host :duration=(:logoff:time - :time))
                ]
            ''')
            self.len(2, nodes)
            self.eq('visi', nodes[0].get('user'))
            self.nn(nodes[0].get('host'))
            self.nn(nodes[0].get('domain'))
            self.nn(nodes[0].get('contact'))

            self.nn(nodes[1].get('host'))
            self.nn(nodes[1].get('account'))
            self.eq(1615680000000, nodes[1].get('time'))
            self.eq(1615687260000, nodes[1].get('logoff:time'))
            self.eq(7260000, nodes[1].get('duration'))
            self.eq('02:01:00.000', nodes[1].repr('duration'))

            nodes = await core.nodes('inet:email=visi@vertex.link -> ps:contact -> it:account -> it:logon +:time>=2021 -> it:host')
            self.len(1, nodes)
            self.eq('it:host', nodes[0].ndef[0])

            nodes = await core.nodes('it:account -> it:domain')
            self.len(1, nodes)
            self.nn(nodes[0].get('org'))
            self.eq('vertex', nodes[0].get('name'))
            self.eq('the vertex project domain', nodes[0].get('desc'))

            nodes = await core.nodes('''[
                it:log:event=*
                    :mesg=foobar
                    :data=(foo, bar, baz)
                    :severity=debug

                    :host={it:host | limit 1}
            ]''')
            self.len(1, nodes)
            self.eq(10, nodes[0].get('severity'))
            self.eq('foobar', nodes[0].get('mesg'))
            self.eq(('foo', 'bar', 'baz'), nodes[0].get('data'))
            # check that the host activity model was inherited
            self.nn(nodes[0].get('host'))

    async def test_it_forms_prodsoft(self):
        # Test all prodsoft and prodsoft associated linked forms
        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                # it:prod:soft
                prod0 = s_common.guid()
                org0 = s_common.guid()
                person0 = s_common.guid()
                file0 = 'a' * 64
                acct0 = ('vertex.link', 'pennywise')
                url0 = 'https://vertex.link/products/balloonmaker'
                sprops = {
                    'name': 'Balloon Maker',
                    'desc': "Pennywise's patented balloon blower upper",
                    'desc:short': 'Balloon blower',
                    'author:org': org0,
                    'author:email': 'pennywise@vertex.link',
                    'author:acct': acct0,
                    'author:person': person0,
                    'url': url0,
                }
                node = await snap.addNode('it:prod:soft', prod0, sprops)
                self.eq(node.ndef[1], prod0)
                self.eq(node.get('name'), 'balloon maker')
                self.eq(node.get('desc'), "Pennywise's patented balloon blower upper")
                self.eq(node.get('desc:short'), 'balloon blower')
                self.eq(node.get('author:org'), org0)
                self.eq(node.get('author:acct'), acct0)
                self.eq(node.get('author:email'), 'pennywise@vertex.link')
                self.eq(node.get('author:person'), person0)
                self.false(node.get('isos'))
                self.false(node.get('islib'))
                await node.set('isos', True)
                await node.set('islib', True)
                self.true(node.get('isos'))
                self.true(node.get('islib'))

                self.eq(node.get('url'), url0)

                # it:prod:softver - this does test a bunch of property related callbacks
                url1 = 'https://vertex.link/products/balloonmaker/release_101-beta.exe'
                vprops = {
                    'vers': 'V1.0.1-beta+exp.sha.5114f85',
                    'released': '2018-04-03 08:44:22',
                    'url': url1,
                    'software': prod0,
                    'arch': 'amd64',
                    'name': 'balloonmaker',
                    'desc': 'makes balloons',
                }
                ver0 = s_common.guid()
                node = await snap.addNode('it:prod:softver', ver0, vprops)

                self.eq(node.ndef[1], ver0)
                self.eq(node.get('arch'), 'amd64')
                self.eq(node.get('released'), 1522745062000)
                self.eq(node.get('software'), prod0)
                self.eq(node.get('software:name'), 'balloon maker')
                self.eq(node.get('vers'), 'V1.0.1-beta+exp.sha.5114f85')
                self.eq(node.get('vers:norm'), 'v1.0.1-beta+exp.sha.5114f85')
                self.eq(node.get('semver'), 0x000010000000001)
                self.eq(node.get('semver:major'), 1)
                self.eq(node.get('semver:minor'), 0)
                self.eq(node.get('semver:patch'), 1)
                self.eq(node.get('semver:pre'), 'beta')
                self.eq(node.get('semver:build'), 'exp.sha.5114f85')
                self.eq(node.get('url'), url1)
                self.eq(node.get('name'), 'balloonmaker')
                self.eq(node.get('desc'), 'makes balloons')
                # callback node creation checks
                nodes = await snap.nodes('it:dev:str=V1.0.1-beta+exp.sha.5114f85')
                self.len(1, nodes)
                nodes = await snap.nodes('it:dev:str=amd64')
                self.len(1, nodes)

                host0 = s_common.guid()
                node = await snap.addNode('it:hostsoft', (host0, ver0))
                self.eq(node.ndef[1], (host0, ver0))
                self.eq(node.get('host'), host0)
                self.eq(node.get('softver'), ver0)

                softfileprops = {'path': '/path/to/nowhere'}
                softfile = await snap.addNode('it:prod:softfile', (ver0, file0), props=softfileprops)
                self.eq(softfile.get('soft'), ver0)
                self.eq(softfile.get('file'), f'sha256:{file0}')
                self.eq('/path/to/nowhere', softfile.get('path'))
                self.len(1, await core.nodes('it:prod:softfile -> file:path'))

                ver1 = s_common.guid()
                softlib = await snap.addNode('it:prod:softlib', (ver0, ver1))
                self.eq(softlib.get('soft'), ver0)
                self.eq(softlib.get('lib'), ver1)

                os0 = s_common.guid()
                softos = await snap.addNode('it:prod:softos', (ver0, os0))
                self.eq(softos.get('soft'), ver0)
                self.eq(softos.get('os'), os0)

                prod1 = s_common.guid()
                sigprops = {
                    'desc': 'The evil balloon virus!',
                    'url': url1,
                }
                sig0 = (prod1, 'Bar.BAZ.faZ')
                node = await snap.addNode('it:av:sig', sig0, sigprops)
                self.eq(node.ndef[1], (prod1, 'Bar.BAZ.faZ'.lower()))
                self.eq(node.get('soft'), prod1)
                self.eq(node.get('name'), 'bar.baz.faz')
                self.eq(node.get('desc'), 'The evil balloon virus!')
                self.eq(node.get('url'), url1)

                node = await snap.addNode('it:av:filehit', (file0, sig0))
                self.eq(node.ndef[1], (f'sha256:{file0}', (prod1, 'Bar.BAZ.faZ'.lower())))
                self.eq(node.get('file'), f'sha256:{file0}')
                self.eq(node.get('sig'), (prod1, 'Bar.BAZ.faZ'.lower()))
                self.eq(node.get('sig:name'), 'bar.baz.faz')
                self.eq(node.get('sig:soft'), prod1)
                await self.checkNodes(core, (('it:prod:soft', prod1),))

                # Test 'vers' semver brute forcing
                testvectors = [
                    ('1', 0x000010000000000, {'major': 1, 'minor': 0, 'patch': 0}),
                    ('2.0A1', 0x000020000000000, {'major': 2, 'minor': 0, 'patch': 0}),
                    ('2016-03-01', 0x007e00000300001, {'major': 2016, 'minor': 3, 'patch': 1}),
                    ('1.2.windows-RC1', 0x000010000200000, {'major': 1, 'minor': 2, 'patch': 0}),
                    ('3.4', 0x000030000400000, {'major': 3, 'minor': 4, 'patch': 0}),
                    ('1.3a2.dev12', 0x000010000000000, {'major': 1, 'minor': 0, 'patch': 0}),
                    ('v2.4.0.0-1', 0x000020000400000, {'major': 2, 'minor': 4, 'patch': 0}),
                    ('v2.4.1.0-0.3.rc1', 0x000020000400001, {'major': 2, 'minor': 4, 'patch': 1}),
                    ('0.18rc2', 0, {'major': 0, 'minor': 0, 'patch': 0}),
                    ('OpenSSL_1_0_2l', 0x000010000000000, {'major': 1, 'minor': 0, 'patch': 0}),
                ]
                for tv, te, subs in testvectors:
                    props = {
                        'vers': tv
                    }
                    node = await snap.addNode('it:prod:softver', '*', props)
                    self.eq(node.get('semver'), te)
                    self.eq(node.get('semver:major'), subs.get('major'))
                    self.eq(node.get('semver:minor'), subs.get('minor'))
                    self.eq(node.get('semver:patch'), subs.get('patch'))

                node = await snap.addNode('it:prod:softver', '*', {'vers': ''})
                self.eq(node.get('vers'), '')
                self.none(node.get('vers:norm'))
                self.none(node.get('semver'))

                with self.getLoggerStream('synapse.models.infotech',
                                          'Unable to brute force version parts out of the string') as stream:

                    node = await snap.addNode('it:prod:softver', '*', {'vers': 'Alpha'})
                    self.none(node.get('semver'))
                    self.true(stream.is_set())

    async def test_it_form_callbacks(self):
        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                # it:dev:str kicks out the :norm property on him when he is made
                node = await snap.addNode('it:dev:str', 'evil RAT')
                self.eq(node.ndef[1], 'evil RAT')
                self.eq(node.get('norm'), 'evil rat')

                node = await snap.addNode('it:dev:pipe', 'MyPipe')
                self.eq(node.ndef[1], 'MyPipe')
                nodes = await snap.nodes('it:dev:str=MyPipe')
                self.len(1, nodes)
                # The callback created node also has norm set on it
                self.eq(nodes[0].get('norm'), 'mypipe')

                node = await snap.addNode('it:dev:mutex', 'MyMutex')
                self.eq(node.ndef[1], 'MyMutex')
                nodes = await snap.nodes('it:dev:str=MyMutex')
                self.len(1, nodes)

                key = 'HKEY_LOCAL_MACHINE\\Foo\\Bar'
                node = await snap.addNode('it:dev:regkey', key)
                self.eq(node.ndef[1], key)
                opts = {'vars': {'key': key}}
                nodes = await snap.nodes('it:dev:str=$key', opts=opts)
                self.len(1, nodes)

                fbyts = 'sha256:' + 64 * 'f'
                key = 'HKEY_LOCAL_MACHINE\\DUCK\\QUACK'
                valus = [
                    ('str', 'knight'),
                    ('int', 20),
                    ('bytes', fbyts),
                ]
                for prop, valu in valus:

                    iden = s_common.guid((key, valu))
                    props = {
                        'key': key,
                        prop: valu,
                    }
                    node = await snap.addNode('it:dev:regval', iden, props)
                    self.eq(node.ndef[1], iden)
                    self.eq(node.get('key'), key)
                    self.eq(node.get(prop), valu)

                nodes = await snap.nodes('it:dev:str=HKEY_LOCAL_MACHINE\\Foo\\Bar')
                self.len(1, nodes)

    async def test_it_semvertype(self):
        async with self.getTestCore() as core:
            t = core.model.type('it:semver')
            testvectors = (
                # Strings
                ('1.2.3', (0x000010000200003,
                           {'major': 1, 'minor': 2, 'patch': 3, })),
                ('0.0.1', (0x000000000000001,
                           {'major': 0, 'minor': 0, 'patch': 1, })),
                ('1.2.3-alpha', (0x000010000200003,
                                 {'major': 1, 'minor': 2, 'patch': 3,
                                  'pre': 'alpha', })),
                ('1.2.3-alpha.1', (0x000010000200003,
                                   {'major': 1, 'minor': 2, 'patch': 3,
                                    'pre': 'alpha.1', })),
                ('1.2.3-0.3.7', (0x000010000200003,
                                 {'major': 1, 'minor': 2, 'patch': 3,
                                  'pre': '0.3.7', })),
                ('1.2.3-x.7.z.92', (0x000010000200003,
                                    {'major': 1, 'minor': 2, 'patch': 3,
                                     'pre': 'x.7.z.92', })),
                ('1.2.3-alpha+001', (0x000010000200003,
                                     {'major': 1, 'minor': 2, 'patch': 3,
                                      'pre': 'alpha', 'build': '001'})),
                ('1.2.3+20130313144700', (0x000010000200003,
                                          {'major': 1, 'minor': 2, 'patch': 3,
                                           'build': '20130313144700'})),
                ('1.2.3-beta+exp.sha.5114f85', (0x000010000200003,
                                                {'major': 1, 'minor': 2, 'patch': 3,
                                                 'pre': 'beta',
                                                 'build': 'exp.sha.5114f85'})),
                # Real world examples
                ('1.2.3-B5CD5743F', (0x000010000200003,
                                     {'major': 1, 'minor': 2, 'patch': 3,
                                      'pre': 'B5CD5743F', })),
                ('V1.2.3', (0x000010000200003,
                            {'major': 1, 'minor': 2, 'patch': 3, })),
                ('V1.4.0-RC0', (0x000010000400000,
                                {'major': 1, 'minor': 4, 'patch': 0,
                                 'pre': 'RC0', })),
                ('v2.4.1-0.3.rc1', (0x000020000400001,
                                    {'major': 2, 'minor': 4, 'patch': 1,
                                     'pre': '0.3.rc1'})),
                ('0.18.1', (0x000000001200001,
                            {'major': 0, 'minor': 18, 'patch': 1, })),
                # Integer values
                (0, (0, {'major': 0, 'minor': 0, 'patch': 0})),
                (1, (1, {'major': 0, 'minor': 0, 'patch': 1})),
                (2, (2, {'major': 0, 'minor': 0, 'patch': 2})),
                (0xFFFFF, (0xFFFFF, {'major': 0, 'minor': 0, 'patch': 0xFFFFF})),
                (0xFFFFF + 1, (0xFFFFF + 1, {'major': 0, 'minor': 1, 'patch': 0})),
                (0xdeadb33f1337133, (0xdeadb33f1337133, {'major': 0xdeadb, 'minor': 0x33f13, 'patch': 0x37133})),
                (0xFFFFFFFFFFFFFFF, (0xFFFFFFFFFFFFFFF, {'major': 0xFFFFF, 'minor': 0xFFFFF, 'patch': 0xFFFFF}))
            )

            for v, e in testvectors:
                ev, es = e
                valu, rdict = t.norm(v)
                subs = rdict.get('subs')
                self.eq(valu, ev)
                self.eq(subs, es)

            testvectors_bad = (
                # Invalid strings
                '1',
                '1.2',
                '2.0A1',
                '0.18rc2',
                '0.0.00001',
                '2016-03-01',
                'v2.4.0.0-1',
                '1.3a2.dev12',
                'OpenSSL_1_0_2l',
                '1.2.windows-RC1',
                'v2.4.1.0-0.3.rc1',
                # invalid ints
                -1,
                0xFFFFFFFFFFFFFFFFFFFFFFFF + 1,
                # Invalid build and prerelease values
                '1.2.3-alpha.foo..+001',
                '1.2.3-alpha.foo.001+001',
                '1.2.3-alpha+001.blahblahblah...',
                '1.2.3-alpha+001.blahblahblah.*iggy',
                # Just bad input
                '   ',
                ' alpha ',
            )
            for v in testvectors_bad:
                self.raises(s_exc.BadTypeValu, t.norm, v)

            testvectors_repr = (
                (0, '0.0.0'),
                (1, '0.0.1'),
                (0x000010000200003, '1.2.3'),
            )
            for v, e in testvectors_repr:
                self.eq(t.repr(v), e)

    async def test_it_forms_screenshot(self):
        async with self.getTestCore() as core:
            nodes = await core.nodes('''[
                it:screenshot=*
                    :host=*
                    :image=*
                    :desc=WootWoot
            ]''')

            self.len(1, nodes)
            self.eq('it:screenshot', nodes[0].ndef[0])
            self.eq('WootWoot', nodes[0].props['desc'])

            self.len(1, await core.nodes('it:screenshot :host -> it:host'))
            self.len(1, await core.nodes('it:screenshot :image -> file:bytes'))

    async def test_it_forms_hardware(self):
        async with self.getTestCore() as core:
            nodes = await core.nodes('''[
                it:prod:hardware=*
                    :make=dell
                    :model=XPS13
                    :version=alpha
                    :type=pc.laptop
                    :desc=WootWoot
                    :released=20220202
                    :cpe=cpe:2.3:h:dell:xps13::::::::
                    :parts = (*, *)
            ]''')
            self.eq('WootWoot', nodes[0].props['desc'])
            self.eq('dell', nodes[0].props['make'])
            self.eq('xps13', nodes[0].props['model'])
            self.eq('alpha', nodes[0].props['version'])
            self.eq('cpe:2.3:h:dell:xps13::::::::', nodes[0].props['cpe'])
            self.eq(1643760000000, nodes[0].props['released'])
            self.len(1, await core.nodes('it:prod:hardware :make -> ou:name'))
            self.len(1, await core.nodes('it:prod:hardware :type -> it:prod:hardwaretype'))
            self.len(2, await core.nodes('it:prod:hardware:make=dell -> it:prod:hardware'))

            nodes = await core.nodes('''[
                it:prod:component=*
                    :hardware={it:prod:hardware:make=dell}
                    :serial=asdf1234
                    :host=*
            ]''')
            self.nn(nodes[0].props['host'])
            self.eq('asdf1234', nodes[0].props['serial'])
            self.len(1, await core.nodes('it:prod:component -> it:host'))
            self.len(1, await core.nodes('it:prod:component -> it:prod:hardware +:make=dell'))

    async def test_it_forms_hostexec(self):
        # forms related to the host execution model
        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                exe = 'sha256:' + 'a' * 64
                port = 80
                tick = s_common.now()
                host = s_common.guid()
                proc = s_common.guid()
                mutex = 'giggleXX_X0'
                pipe = 'pipe\\mynamedpipe'
                user = 'serviceadmin'
                pid = 20
                key = 'HKEY_LOCAL_MACHINE\\Foo\\Bar'
                ipv4 = 0x01020304
                ipv6 = '::1'

                addr4 = f'tcp://1.2.3.4:{port}'
                addr6 = f'udp://[::1]:{port}'
                url = 'http://www.google.com/sekrit.html'
                raw_path = r'c:\Windows\System32\rar.exe'
                norm_path = r'c:/windows/system32/rar.exe'
                src_proc = s_common.guid()
                src_path = r'c:/temp/ping.exe'
                cmd0 = 'rar a -r yourfiles.rar *.txt'
                fpath = 'c:/temp/yourfiles.rar'
                fbyts = 'sha256:' + 'b' * 64
                pprops = {
                    'exe': exe,
                    'pid': pid,
                    'cmd': cmd0,
                    'host': host,
                    'time': tick,
                    'user': user,
                    'account': '*',
                    'path': raw_path,
                    'src:exe': src_path,
                    'src:proc': src_proc,
                }
                node = await snap.addNode('it:exec:proc', proc, pprops)
                self.eq(node.ndef[1], proc)
                self.eq(node.get('exe'), exe)
                self.eq(node.get('pid'), pid)
                self.eq(node.get('cmd'), cmd0)
                self.eq(node.get('host'), host)
                self.eq(node.get('time'), tick)
                self.eq(node.get('user'), user)
                self.eq(node.get('path'), norm_path)
                self.eq(node.get('src:exe'), src_path)
                self.eq(node.get('src:proc'), src_proc)

                self.nn(node.get('account'))
                self.len(1, await core.nodes('it:exec:proc -> it:account'))

                nodes = await core.nodes('it:cmd')
                self.len(1, nodes)
                self.eq(nodes[0].ndef, ('it:cmd', 'rar a -r yourfiles.rar *.txt'))

                m0 = s_common.guid()
                mprops = {
                    'exe': exe,
                    'proc': proc,
                    'name': mutex,
                    'host': host,
                    'time': tick,
                }
                node = await snap.addNode('it:exec:mutex', m0, mprops)
                self.eq(node.ndef[1], m0)
                self.eq(node.get('exe'), exe)
                self.eq(node.get('proc'), proc)
                self.eq(node.get('host'), host)
                self.eq(node.get('time'), tick)
                self.eq(node.get('name'), mutex)

                p0 = s_common.guid()
                pipeprops = {
                    'exe': exe,
                    'proc': proc,
                    'name': pipe,
                    'host': host,
                    'time': tick,
                }
                node = await snap.addNode('it:exec:pipe', p0, pipeprops)
                self.eq(node.ndef[1], p0)
                self.eq(node.get('exe'), exe)
                self.eq(node.get('proc'), proc)
                self.eq(node.get('host'), host)
                self.eq(node.get('time'), tick)
                self.eq(node.get('name'), pipe)

                u0 = s_common.guid()
                uprops = {
                    'proc': proc,
                    'host': host,
                    'exe': exe,
                    'time': tick,
                    'url': url,
                    'page:pdf': '*',
                    'page:html': '*',
                    'page:image': '*',
                    'browser': '*',
                    'client': addr4,
                }
                node = await snap.addNode('it:exec:url', u0, uprops)
                self.eq(node.ndef[1], u0)
                self.eq(node.get('exe'), exe)
                self.eq(node.get('proc'), proc)
                self.eq(node.get('host'), host)
                self.eq(node.get('time'), tick)
                self.eq(node.get('url'), url)
                self.eq(node.get('client'), addr4)
                self.eq(node.get('client:ipv4'), ipv4)
                self.eq(node.get('client:port'), port)

                self.nn(node.get('page:pdf'))
                self.nn(node.get('page:html'))
                self.nn(node.get('page:image'))
                self.nn(node.get('browser'))
                opts = {'vars': {'guid': u0}}
                self.len(1, await core.nodes('it:exec:url=$guid :page:pdf -> file:bytes', opts=opts))
                self.len(1, await core.nodes('it:exec:url=$guid :page:html -> file:bytes', opts=opts))
                self.len(1, await core.nodes('it:exec:url=$guid :page:image -> file:bytes', opts=opts))
                self.len(1, await core.nodes('it:exec:url=$guid :browser -> it:prod:softver', opts=opts))

                u1 = s_common.guid()
                uprops['client'] = addr6
                node = await snap.addNode('it:exec:url', u1, uprops)
                self.eq(node.ndef[1], u1)
                self.eq(node.get('client'), addr6)
                self.eq(node.get('client:ipv6'), ipv6)
                self.eq(node.get('client:port'), port)

                b0 = s_common.guid()
                bprops = {
                    'proc': proc,
                    'host': host,
                    'exe': exe,
                    'time': tick,
                    'server': addr4
                }
                node = await snap.addNode('it:exec:bind', b0, bprops)
                self.eq(node.ndef[1], b0)
                self.eq(node.get('exe'), exe)
                self.eq(node.get('proc'), proc)
                self.eq(node.get('host'), host)
                self.eq(node.get('time'), tick)
                self.eq(node.get('server'), addr4)
                self.eq(node.get('server:ipv4'), ipv4)
                self.eq(node.get('server:port'), port)

                b1 = s_common.guid()
                bprops['server'] = addr6
                node = await snap.addNode('it:exec:bind', b1, bprops)
                self.eq(node.ndef[1], b1)
                self.eq(node.get('server'), addr6)
                self.eq(node.get('server:ipv6'), ipv6)
                self.eq(node.get('server:port'), port)

                faprops = {
                    'exe': exe,
                    'host': host,
                    'proc': proc,
                    'file': fbyts,
                    'time': tick,
                    'path': fpath,
                }
                fa0 = s_common.guid()
                node = await snap.addNode('it:exec:file:add', fa0, faprops)
                self.eq(node.ndef[1], fa0)
                self.eq(node.get('exe'), exe)
                self.eq(node.get('host'), host)
                self.eq(node.get('proc'), proc)
                self.eq(node.get('time'), tick)
                self.eq(node.get('file'), fbyts)
                self.eq(node.get('path'), fpath)
                self.eq(node.get('path:dir'), 'c:/temp')
                self.eq(node.get('path:base'), 'yourfiles.rar')
                self.eq(node.get('path:ext'), 'rar')

                fr0 = s_common.guid()
                node = await snap.addNode('it:exec:file:read', fr0, faprops)
                self.eq(node.ndef[1], fr0)
                self.eq(node.get('exe'), exe)
                self.eq(node.get('host'), host)
                self.eq(node.get('proc'), proc)
                self.eq(node.get('time'), tick)
                self.eq(node.get('file'), fbyts)
                self.eq(node.get('path'), fpath)
                self.eq(node.get('path:dir'), 'c:/temp')
                self.eq(node.get('path:base'), 'yourfiles.rar')
                self.eq(node.get('path:ext'), 'rar')

                fw0 = s_common.guid()
                node = await snap.addNode('it:exec:file:write', fw0, faprops)
                self.eq(node.ndef[1], fw0)
                self.eq(node.get('exe'), exe)
                self.eq(node.get('host'), host)
                self.eq(node.get('proc'), proc)
                self.eq(node.get('time'), tick)
                self.eq(node.get('file'), fbyts)
                self.eq(node.get('path'), fpath)
                self.eq(node.get('path:dir'), 'c:/temp')
                self.eq(node.get('path:base'), 'yourfiles.rar')
                self.eq(node.get('path:ext'), 'rar')

                fd0 = s_common.guid()
                node = await snap.addNode('it:exec:file:del', fd0, faprops)
                self.eq(node.ndef[1], fd0)
                self.eq(node.get('exe'), exe)
                self.eq(node.get('host'), host)
                self.eq(node.get('proc'), proc)
                self.eq(node.get('time'), tick)
                self.eq(node.get('file'), fbyts)
                self.eq(node.get('path'), fpath)
                self.eq(node.get('path:dir'), 'c:/temp')
                self.eq(node.get('path:base'), 'yourfiles.rar')
                self.eq(node.get('path:ext'), 'rar')

                file0 = s_common.guid()
                fsprops = {
                    'host': host,
                    'path': fpath,
                    'file': fbyts,
                    'ctime': tick,
                    'mtime': tick + 1,
                    'atime': tick + 2,
                    'user': user,
                    'group': 'domainadmin'
                }
                node = await snap.addNode('it:fs:file', file0, fsprops)
                self.eq(node.ndef[1], file0)
                self.eq(node.get('host'), host)
                self.eq(node.get('user'), user)
                self.eq(node.get('group'), 'domainadmin')
                self.eq(node.get('file'), fbyts)
                self.eq(node.get('ctime'), tick)
                self.eq(node.get('mtime'), tick + 1)
                self.eq(node.get('atime'), tick + 2)
                self.eq(node.get('path'), fpath)
                self.eq(node.get('path:dir'), 'c:/temp')
                self.eq(node.get('path:base'), 'yourfiles.rar')
                self.eq(node.get('path:ext'), 'rar')

                rprops = {
                    'host': host,
                    'proc': proc,
                    'exe': exe,
                    'time': tick,
                    'reg': '*',
                }
                forms = ('it:exec:reg:get',
                         'it:exec:reg:set',
                         'it:exec:reg:del',
                         )
                for form in forms:
                    rk0 = s_common.guid()
                    nprops = rprops.copy()
                    node = await snap.addNode(form, rk0, nprops)
                    self.eq(node.ndef[1], rk0)
                    self.eq(node.get('host'), host)
                    self.eq(node.get('proc'), proc)
                    self.eq(node.get('exe'), exe)
                    self.eq(node.get('time'), tick)
                    self.nn(node.get('reg'))

    async def test_it_app_yara(self):

        async with self.getTestCore() as core:

            rule = s_common.guid()
            opts = {'vars': {'rule': rule}}

            nodes = await core.nodes('[ it:app:yara:rule=$rule :enabled=true :text=gronk :author=* :name=foo :version=1.2.3 ]', opts=opts)

            self.len(1, nodes)
            self.eq('foo', nodes[0].get('name'))
            self.eq(True, nodes[0].get('enabled'))
            self.eq('gronk', nodes[0].get('text'))
            self.eq(0x10000200003, nodes[0].get('version'))

            self.len(1, await core.nodes('it:app:yara:rule=$rule -> ps:contact', opts=opts))

            nodes = await core.nodes('[ it:app:yara:match=($rule, "*") :version=1.2.3 ]', opts=opts)
            self.len(1, nodes)
            self.nn(nodes[0].get('file'))
            self.eq(rule, nodes[0].get('rule'))
            self.eq(0x10000200003, nodes[0].get('version'))

    async def test_it_app_snort(self):

        async with self.getTestCore() as core:

            hit = s_common.guid()
            rule = s_common.guid()
            flow = s_common.guid()
            host = s_common.guid()
            opts = {'vars': {'rule': rule, 'flow': flow, 'host': host, 'hit': hit}}

            nodes = await core.nodes('[ it:app:snort:rule=$rule :text=gronk :name=foo :version=1.2.3 ]', opts=opts)

            self.len(1, nodes)
            self.eq('foo', nodes[0].get('name'))
            self.eq('gronk', nodes[0].get('text'))
            self.eq(0x10000200003, nodes[0].get('version'))

            nodes = await core.nodes('[ it:app:snort:hit=$hit :rule=$rule :flow=$flow :src="tcp://[::ffff:0102:0304]:0" :dst="tcp://[::ffff:0505:0505]:80" :time=2015 :sensor=$host :version=1.2.3 ]', opts=opts)
            self.len(1, nodes)
            self.eq(rule, nodes[0].get('rule'))
            self.eq(flow, nodes[0].get('flow'))
            self.eq(host, nodes[0].get('sensor'))
            self.eq(1420070400000, nodes[0].get('time'))

            self.eq('tcp://[::ffff:1.2.3.4]:0', nodes[0].get('src'))
            self.eq(0, nodes[0].get('src:port'))
            self.eq(0x01020304, nodes[0].get('src:ipv4'))
            self.eq('::ffff:1.2.3.4', nodes[0].get('src:ipv6'))

            self.eq('tcp://[::ffff:5.5.5.5]:80', nodes[0].get('dst'))
            self.eq(80, nodes[0].get('dst:port'))
            self.eq(0x05050505, nodes[0].get('dst:ipv4'))
            self.eq('::ffff:5.5.5.5', nodes[0].get('dst:ipv6'))

            self.eq(0x10000200003, nodes[0].get('version'))

    async def test_it_reveng(self):

        async with self.getTestCore() as core:

            baseFile = s_common.ehex(s_common.buid())
            func = s_common.guid()
            fva = 0x404438
            rank = 33
            complexity = 60
            funccalls = ((baseFile, func), )
            fopt = {'vars': {'file': baseFile,
                             'func': func,
                             'fva': fva,
                             'rank': rank,
                             'cmplx': complexity,
                             'funccalls': funccalls}}
            vstr = 'VertexBrandArtisanalBinaries'
            sopt = {'vars': {'func': func,
                             'string': vstr}}
            name = "FunkyFunction"
            descrp = "Test Function"
            impcalls = ("libr.foo", "libr.foo2", "libr.foo3")
            funcopt = {'vars': {'name': name,
                                'descrp': descrp,
                                'impcalls': impcalls}}

            fnode = await core.nodes('[it:reveng:filefunc=($file, $func) :va=$fva :rank=$rank :complexity=$cmplx :funccalls=$funccalls]', opts=fopt)
            snode = await core.nodes('[it:reveng:funcstr=($func, $string)]', opts=sopt)
            self.len(1, fnode)
            self.eq(f'sha256:{baseFile}', fnode[0].get('file'))
            self.eq(fva, fnode[0].get('va'))
            self.eq(rank, fnode[0].get('rank'))
            self.eq(complexity, fnode[0].get('complexity'))
            self.eq((f'sha256:{baseFile}', func), fnode[0].get('funccalls')[0])

            self.len(1, snode)
            self.eq(fnode[0].get('function'), snode[0].get('function'))
            self.eq(vstr, snode[0].get('string'))

            funcnode = await core.nodes('''
                it:reveng:function [
                    :name=$name
                    :description=$descrp
                    :impcalls=$impcalls
                    :strings=(bar,foo,foo)
            ]''', opts=funcopt)
            self.len(1, funcnode)
            self.eq(name, funcnode[0].get('name'))
            self.eq(descrp, funcnode[0].get('description'))
            self.len(len(impcalls), funcnode[0].get('impcalls'))
            self.eq(impcalls[0], funcnode[0].get('impcalls')[0])
            self.sorteq(('bar', 'foo'), funcnode[0].get('strings'))

            nodes = await core.nodes('it:reveng:function -> it:dev:str')
            self.len(2, nodes)

            nodes = await core.nodes(f'file:bytes={baseFile} -> it:reveng:filefunc :function -> it:reveng:funcstr:function')
            self.len(1, nodes)
            self.eq(vstr, nodes[0].get('string'))

            nodes = await core.nodes(f'file:bytes={baseFile} -> it:reveng:filefunc -> it:reveng:function -> it:reveng:impfunc')
            self.len(len(impcalls), nodes)

    async def test_infotech_cpes(self):

        async with self.getTestCore() as core:

            cpe23 = core.model.type('it:sec:cpe')
            cpe22 = core.model.type('it:sec:cpe:v2_2')

            with self.raises(s_exc.BadTypeValu):
                cpe22.norm('cpe:/a:vertex:synapse:0:1:2:3:4:5:6:7:8:9')

            with self.raises(s_exc.BadTypeValu):
                cpe23.norm('cpe:/a:vertex:synapse:0:1:2:3:4:5:6:7:8:9')

            with self.raises(s_exc.BadTypeValu):
                cpe23.norm('cpe:2.3:a:vertex:synapse')

            # test cast 2.2 -> 2.3 upsample
            norm, info = cpe23.norm('cpe:/a:vertex:synapse')
            self.eq(norm, 'cpe:2.3:a:vertex:synapse:*:*:*:*:*:*:*:*')

            # test cast 2.3 -> 2.2 downsample
            norm, info = cpe22.norm('cpe:2.3:a:vertex:synapse:*:*:*:*:*:*:*:*')
            self.eq(norm, 'cpe:/a:vertex:synapse')

            nodes = await core.nodes('[ it:sec:cpe=cpe:2.3:a:vertex:synapse:*:*:*:*:*:*:*:* ]')
            self.eq('cpe:/a:vertex:synapse', nodes[0].props['v2_2'])

            # test lift by either via upsample and downsample
            self.len(1, await core.nodes('it:sec:cpe=cpe:/a:vertex:synapse +:v2_2=cpe:/a:vertex:synapse'))
            self.len(1, await core.nodes('it:sec:cpe=cpe:2.3:a:vertex:synapse:*:*:*:*:*:*:*:*'))
            self.len(1, await core.nodes('it:sec:cpe:v2_2=cpe:/a:vertex:synapse'))
            self.len(1, await core.nodes('it:sec:cpe:v2_2=cpe:2.3:a:vertex:synapse:*:*:*:*:*:*:*:*'))



import binascii

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.stormtypes as s_stormtypes

@s_stormtypes.registry.registerLib
class HexLib(s_stormtypes.Lib):
    '''
    A Storm library which implements helpers for hexadecimal encoded strings.
    '''
    _storm_locals = (
        {'name': 'encode', 'desc': 'Encode bytes into a hexadecimal string.',
         'type': {'type': 'function', '_funcname': 'encode',
                  'args': (
                      {'name': 'valu', 'type': 'bytes', 'desc': 'The bytes to be encoded into a hex string.'},
                  ),
                  'returns': {'type': 'str', 'desc': 'The hex encoded string.', }
        }},
        {'name': 'decode', 'desc': 'Decode a hexadecimal string into bytes.',
         'type': {'type': 'function', '_funcname': 'decode',
                  'args': (
                      {'name': 'valu', 'type': 'str', 'desc': 'The hex string to be decoded into bytes.'},
                  ),
                  'returns': {'type': 'bytes', 'desc': 'The decoded bytes.', }
        }},
    )

    _storm_lib_path = ('hex',)

    def getObjLocals(self):
        return {
            # TODO 'dump': self.dump,
            'encode': self.encode,
            'decode': self.decode,
        }

    async def encode(self, valu):
        if not isinstance(valu, bytes):
            raise s_exc.BadArg(mesg='$lib.hex.encode() requires a bytes argument.')
        return s_common.ehex(valu)

    async def decode(self, valu):
        valu = await s_stormtypes.tostr(valu)
        try:
            return s_common.uhex(valu)
        except binascii.Error as e:
            raise s_exc.BadArg(mesg=f'$lib.hex.decode(): {e}')



import synapse.exc as s_exc

import synapse.tests.utils as s_test

class StormIpv6Test(s_test.SynTest):

    async def test_storm_ipv6(self):
        async with self.getTestCore() as core:

            self.len(1, await core.nodes('[inet:ipv6=2001:4860:4860::8888]'))
            query = 'inet:ipv6=2001:4860:4860::8888 return ( $lib.inet.ipv6.expand($node.value()) )'
            self.eq('2001:4860:4860:0000:0000:0000:0000:8888',
                    await core.callStorm(query))

            query = '$valu="2001:4860:4860::8888 " return ( $lib.inet.ipv6.expand($valu) )'
            self.eq('2001:4860:4860:0000:0000:0000:0000:8888',
                    await core.callStorm(query))

            query = '$valu="2001:4860:4860::XXXX" return ( $lib.inet.ipv6.expand($valu) )'
            with self.raises(s_exc.StormRuntimeError):
                await core.callStorm(query)



import os
import sys
import signal
import asyncio
import contextlib
import multiprocessing

import synapse.exc as s_exc

import synapse.lib.base as s_base
import synapse.lib.coro as s_coro

import synapse.tests.utils as s_t_utils

def block_processing(evt1):
    '''
    Function to make a base and call main().  Used as a Process target.

    Args:
        evt1 (multiprocessing.Event): event to twiddle
    '''
    async def main():

        base = await s_base.Base.anit()
        await base.addSignalHandlers()

        evt1.set()

        await base.waitfini()

    asyncio.run(main())
    sys.exit(137)

class Hehe(s_base.Base):

    async def __anit__(self, foo):
        await s_base.Base.__anit__(self)
        self.foo = foo
        self.bar = self.foo + 10
        self.posted = False

    async def postAnit(self):
        self.posted = True
        if self.foo == -1:
            raise s_exc.BadArg(mesg='boom')

class BaseTest(s_t_utils.SynTest):

    async def test_base_basics(self):
        base = await s_base.Base.anit()

        def foo(event):
            x = event[1].get('x')
            y = event[1].get('y')
            event[1]['ret'] = x + y

        base.on('woot', foo)

        event = await base.fire('woot', x=3, y=5, ret=[])
        self.eq(event[1]['ret'], 8)

    async def test_base_anit(self):

        afoo = await Hehe.anit(20)
        self.eq(afoo.foo, 20)
        self.eq(afoo.bar, 30)
        self.true(afoo.posted)

        with self.raises(s_exc.BadArg) as cm:
            await Hehe.anit(-1)
        self.eq(cm.exception.get('mesg'), 'boom')

    async def test_coro_fini(self):

        event = asyncio.Event()

        async def setit():
            event.set()

        f = await s_base.Base.anit()
        async with f as f:
            f.onfini(setit)

        self.true(f.isfini)
        self.true(event.is_set())

    async def test_enter_context(self):
        state = None

        @contextlib.contextmanager
        def ctxtest():
            nonlocal state
            state = "before"
            yield state
            state = "after"

        @contextlib.asynccontextmanager
        async def actxtest():
            nonlocal state
            state = "before2"
            yield state
            state = "after2"

        async with await s_base.Base.anit() as base:
            await base.enter_context(ctxtest())
            self.eq("before", state)

        self.eq("after", state)

        async with await s_base.Base.anit() as base:
            await base.enter_context(actxtest())
            self.eq("before2", state)

        self.eq("after2", state)

    async def test_base_link(self):

        base1 = await s_base.Base.anit()
        base2 = await s_base.Base.anit()

        base1.link(base2.dist)

        data = {}

        async def woot(event):
            data['woot'] = True

        base2.on('woot', woot)

        await base1.fire('woot')

        self.true(data.get('woot'))

    async def test_base_unlink(self):

        base = await s_base.Base.anit()

        mesgs = []

        async def woot(mesg):
            mesgs.append(mesg)

        base.link(woot)

        await base.fire('haha')
        self.eq(len(mesgs), 1)

        base.unlink(woot)

        await base.fire('haha')
        self.eq(len(mesgs), 1)

        await base.fini()

    async def test_base_withfini(self):

        data = {'count': 0}

        def onfini():
            data['count'] += 1

        async with await s_base.Base.anit() as base:
            base.onfini(onfini)

        self.eq(data['count'], 1)

    async def test_base_finionce(self):

        data = {'count': 0}

        async def onfini():
            data['count'] += 1

        base = await s_base.Base.anit()
        base.onfini(onfini)

        await base.fini()
        await base.fini()

        self.eq(data['count'], 1)

    async def test_base_off(self):
        base = await s_base.Base.anit()

        data = {'count': 0}

        async def woot(mesg):
            data['count'] += 1

        base.on('hehe', woot)

        await base.fire('hehe')

        base.off('hehe', woot)

        await base.fire('hehe')

        await base.fini()

        self.eq(data['count'], 1)

    async def test_base_waiter(self):
        base0 = await s_base.Base.anit()

        wait0 = base0.waiter(3, 'foo:bar')

        await base0.fire('foo:bar')
        await base0.fire('foo:bar')
        await base0.fire('foo:bar')

        evts = await wait0.wait(timeout=3)
        self.eq(len(evts), 3)

        wait1 = base0.waiter(3, 'foo:baz')
        evts = await wait1.wait(timeout=0.1)
        self.none(evts)

        # Bare waiter test - uses the link() method on the Base
        # to receive all of the events from the Base.
        wait2 = s_base.Waiter(base0, 2)
        await base0.fire('hehe')
        await base0.fire('haha')
        evts = await wait2.wait(1)
        self.len(2, evts)

    async def test_baseref(self):

        bref = await s_base.BaseRef.anit()

        base0 = await s_base.Base.anit()
        base1 = await s_base.Base.anit()
        base2 = await s_base.Base.anit()

        bref.put('foo', base0)
        bref.put('bar', base1)
        bref.put('baz', base2)

        await base1.fini()
        self.nn(bref.get('foo'))
        self.none(bref.get('bar'))

        self.len(2, list(bref))

        self.true(bref.pop('baz') is base2)
        self.len(1, list(bref))

        await bref.fini()
        self.true(base0.isfini)

        await base2.fini()

    async def test_base_waitfini(self):
        loop = asyncio.get_running_loop()

        base = await s_base.Base.anit()

        self.false(await base.waitfini(timeout=0.1))

        async def callfini():
            await asyncio.sleep(0.1)
            await base.fini()

        loop.create_task(callfini())
        # actually wait...
        self.true(await base.waitfini(timeout=0.3))
        self.true(base.isfini)

        # bounce off the isfini block
        self.true(await base.waitfini(timeout=0.3))

    async def test_base_refcount(self):
        base = await s_base.Base.anit()

        self.eq(base.incref(), 2)

        self.eq(await base.fini(), 1)
        self.false(base.isfini)

        self.eq(await base.fini(), 0)
        self.true(base.isfini)

    async def test_baseref_gen(self):

        async with await s_base.BaseRef.anit() as refs:
            await self.asyncraises(s_exc.NoSuchCtor, refs.gen('woot'))

        async def ctor(name):
            return await s_base.Base.anit()

        async with await s_base.BaseRef.anit(ctor=ctor) as refs:

            self.none(refs.get('woot'))

            woot = await refs.gen('woot')
            self.eq(1, woot._syn_refs)

            self.nn(woot)
            self.true(await refs.gen('woot') is woot)
            self.eq(2, woot._syn_refs)

            await woot.fini()
            self.false(woot.isfini)
            self.true(refs.get('woot') is woot)
            self.eq(1, woot._syn_refs)

            await woot.fini()
            self.eq(0, woot._syn_refs)

            self.true(woot.isfini)
            self.false(refs.get('woot') is woot)
            self.eq(0, woot._syn_refs)

    async def test_baseref_items(self):

        bref = await s_base.BaseRef.anit()

        base0 = await s_base.Base.anit()
        base1 = await s_base.Base.anit()
        base2 = await s_base.Base.anit()

        bref.put('foo', base0)
        bref.put('bar', base1)
        bref.put('baz', base2)

        items = bref.items()
        self.isin(('foo', base0), items)
        self.isin(('bar', base1), items)
        self.isin(('baz', base2), items)

        await base1.fini()
        items = bref.items()
        self.isin(('foo', base0), items)
        self.isin(('baz', base2), items)

        await base2.fini()
        items = bref.items()
        self.isin(('foo', base0), items)

        await base0.fini()
        items = bref.items()
        self.eq(items, [])

        await bref.fini()
        items = bref.items()
        self.eq(items, [])

    def test_base_main_sigterm(self):
        self.thisHostMustNot(platform='windows')
        # We have no reliable way to test this on windows
        self.thisHostMustNot(platform='darwin')
        # This test fails on darwin in circleci

        ctx = multiprocessing.get_context('spawn')

        evt1 = ctx.Event()

        proc = ctx.Process(target=block_processing, args=(evt1,))
        proc.start()

        self.true(evt1.wait(timeout=10))
        os.kill(proc.pid, signal.SIGTERM)
        proc.join(timeout=10)
        self.eq(proc.exitcode, 137)

    def test_base_main_sigint(self):
        self.thisHostMustNot(platform='windows')
        # We have no reliable way to test this on windows
        self.thisHostMustNot(platform='darwin')
        # This test fails on darwin in circleci

        ctx = multiprocessing.get_context('spawn')

        evt1 = ctx.Event()

        proc = ctx.Process(target=block_processing, args=(evt1,))
        proc.start()

        self.true(evt1.wait(timeout=10))
        os.kill(proc.pid, signal.SIGINT)

        proc.join(timeout=10)
        self.eq(proc.exitcode, 137)

    async def test_onwith(self):
        base = await s_base.Base.anit()
        l0 = []
        l1 = []

        def onHehe0(mesg):
            l0.append(mesg)

        def onHehe1(mesg):
            l1.append(mesg)

        base.on('hehe', onHehe0)

        # Temporarily set the 'hehe' callback
        with base.onWith('hehe', onHehe1) as e:
            self.true(e is base)
            await base.fire('hehe')
            self.len(1, l0)
            self.len(1, l1)

        # subsequent fires do not call onHehe1
        await base.fire('hehe')
        self.len(2, l0)
        self.len(1, l1)

    async def test_base_mixin(self):

        data = []

        class M1(s_base.Base):
            async def __anit__(self):
                await s_base.Base.__anit__(self)
                self.m1fini = asyncio.Event()
                self.onfini(self.m1fini.set)
                self.on('event', self._M1OnEvent)

            def _M1OnEvent(self, event):
                data.append(event)

        class M2(s_base.Base):
            async def __anit__(self):
                await s_base.Base.__anit__(self)
                self.m2fini = asyncio.Event()
                self.onfini(self.m2fini.set)
                self.on('event', self._M2OnEvent)

            def _M2OnEvent(self, event):
                data.append(event)

        class MixedBases(M1, M2):

            async def __anit__(self):
                # Initialize our mixins
                await M1.__anit__(self)
                await M2.__anit__(self)

        mixed = await MixedBases.anit()
        self.false(mixed.m1fini.is_set())
        self.false(mixed.m2fini.is_set())
        self.len(2, mixed._fini_funcs)
        self.len(2, mixed._syn_funcs.get('event'))
        self.eq(mixed._syn_refs, 1)

        await mixed.fire('event', key=1)
        self.len(2, data)

        await mixed.fini()

        self.true(mixed.m1fini.is_set())
        self.true(mixed.m2fini.is_set())

    async def test_base_schedgenr(self):
        async def goodgenr():
            yield 'foo'
            await asyncio.sleep(0)
            yield 'bar'
            await asyncio.sleep(0)

        async def badgenr():
            yield 'foo'
            await asyncio.sleep(0)
            raise s_exc.SynErr('rando')

        async def slowgenr():
            yield 'foo'
            await asyncio.sleep(5)
            yield 'bar'

        event = asyncio.Event()

        async def slowtask():
            g = s_base.schedGenr(slowgenr())
            item = await g.__anext__()
            self.eq('foo', item)
            event.set()
            item = await g.__anext__()

        items = [item async for item in s_base.schedGenr(goodgenr())]
        self.eq(items, ['foo', 'bar'])
        await self.agenraises(s_exc.SynErr, s_base.schedGenr(badgenr()))

        async with await s_base.Base.anit() as base:
            task = base.schedCoro(slowtask())
            self.true(await s_coro.event_wait(event, 3.0))

        await self.asyncraises(asyncio.CancelledError, task)



# Mach-O Load Command Types
loadcmd_types = (
    (1, 'segment'),
    (2, 'symbol table'),
    (3, 'gdb symbol table'),
    (4, 'thread'),
    (5, 'unix thread'),
    (6, 'fixed VM shared library'),
    (7, 'fixed VM shared library identification'),
    (8, 'object identification'),
    (9, 'fixed VM file inclusion'),
    (10, 'prepage'),
    (11, 'dynamic link-edit symbol table'),
    (12, 'load dynamically linked shared library'),
    (13, 'dynamically linked shared library identifier'),
    (14, 'load dynamic linker'),
    (15, 'dynamic linker identification'),
    (16, 'prebound dynamically linked shared library'),
    (17, 'image routines'),
    (18, 'sub framework'),
    (19, 'sub umbrella'),
    (20, 'sub client'),
    (21, 'sub library'),
    (22, 'two level namespace lookup hints'),
    (23, 'prebind checksum'),
    (24, 'weak import dynamically linked shared library'),
    (25, '64bit segment'),
    (26, '64bit image routines'),
    (27, 'uuid'),
    (28, 'runpath additions'),
    (29, 'code signature'),
    (30, 'split segment info'),
    (31, 'load and re-export dynamic library'),
    (32, 'delay load of dynamic library'),
    (33, 'encrypted segment information'),
    (34, 'compressed dynamic library information'),
    (35, 'load upward dylib'),
    (36, 'minimum osx version'),
    (37, 'minimum ios version'),
    (38, 'compressed table of function start addresses'),
    (39, 'environment variable string for dynamic library'),
    (40, 'unix thread replacement'),
    (41, 'table of non-instructions in __text'),
    (42, 'source version used to build binary'),
    (43, 'Code signing DRs copied from linked dynamic libraries'),
)

def getLoadCmdTypes():
    return loadcmd_types

# Mach-0 section types
section_types = (
    (0, 'regular'),
    (1, 'zero fill on demand'),
    (2, 'only literal C strings'),
    (3, 'only 4 byte literals'),
    (4, 'only 8 byte literals'),
    (5, 'only pointers to literals'),
    (6, 'only non-lazy symbol pointers'),
    (7, 'only lazy symbol pointers'),
    (8, 'only symbol stubs'),
    (9, 'only function pointers for init'),
    (10, 'only function pointers for fini'),
    (11, 'contains symbols to be coalesced'),
    (12, 'zero fill on deman (greater than 4gb)'),
    (13, 'only pairs of function pointers for interposing'),
    (14, 'only 16 byte literals'),
    (15, 'dtrace object format'),
    (16, 'only lazy symbols pointers to lazy dynamic libraries'),
)

def getSectionTypes():
    return section_types



import synapse.tests.utils as s_t_utils

class LangModuleTest(s_t_utils.SynTest):

    async def test_model_language(self):

        async with self.getTestCore() as core:
            nodes = await core.nodes('''[
                lang:translation=*
                    :input=Hola
                    :input:lang=ES
                    :output=Hi
                    :output:lang=en.us
                    :desc=Greetings
                    :engine=*
            ]''')
            self.len(1, nodes)
            self.eq('Hola', nodes[0].get('input'))
            self.eq('Hi', nodes[0].get('output'))
            self.eq('es', nodes[0].get('input:lang'))
            self.eq('en.us', nodes[0].get('output:lang'))
            self.eq('Greetings', nodes[0].get('desc'))
            self.len(1, await core.nodes('lang:translation -> it:prod:softver'))

    async def test_forms_idiom(self):
        async with self.getTestCore() as core:
            formname = 'lang:idiom'
            valu = 'arbitrary text 123'

            input_props = {'url': 'https://vertex.link/', 'desc:en': 'Some English Desc'}
            expected_props = {'url': 'https://vertex.link/', 'desc:en': 'Some English Desc'}
            expected_ndef = (formname, valu)

            async with await core.snap() as snap:
                node = await snap.addNode(formname, valu, props=input_props)

            self.eq(node.ndef, expected_ndef)
            for prop, valu in expected_props.items():
                self.eq(node.get(prop), valu)

    async def test_forms_trans(self):
        async with self.getTestCore() as core:
            formname = 'lang:trans'
            valu = 'arbitrary text 123'

            input_props = {'text:en': 'Some English Text', 'desc:en': 'Some English Desc'}
            expected_props = {'text:en': 'Some English Text', 'desc:en': 'Some English Desc'}
            expected_ndef = (formname, valu)

            async with await core.snap() as snap:
                node = await snap.addNode(formname, valu, props=input_props)

            self.eq(node.ndef, expected_ndef)
            for prop, valu in expected_props.items():
                self.eq(node.get(prop), valu)

    async def test_types_unextended(self):
        # The following types are subtypes that do not extend their base type
        async with self.getTestCore() as core:
            self.nn(core.model.type('lang:idiom'))  # str
            self.nn(core.model.type('lang:trans'))  # str



import synapse.lib.storm_format as s_storm_format
import synapse.lib.parser as s_parser

from synapse.tests.test_lib_grammar import Queries

def test_highlight_storm():

    for _, query in enumerate(Queries):
        s_storm_format.highlight_storm(s_parser.LarkParser, query)



import hashlib

import synapse.axon as s_axon

import synapse.tests.utils as s_t_utils

asdfhash = hashlib.sha256(b'asdfasdf').digest()

class AxonServerTest(s_t_utils.SynTest):

    async def test_server(self):

        with self.getTestDir() as dirn:
            async with self.withSetLoggingMock() as mock:

                argv = [dirn,
                        '--telepath', 'tcp://127.0.0.1:0/',
                        '--https', '0',
                        '--name', 'teleaxon']

                async with await s_axon.Axon.initFromArgv(argv,) as axon:
                    async with axon.getLocalProxy() as proxy:
                        async with await proxy.upload() as fd:
                            await fd.write(b'asdfasdf')
                            await fd.save()

                    self.true(axon.dmon.shared.get('teleaxon') is axon)

                # And data persists...
                async with await s_axon.Axon.initFromArgv(argv,) as axon:
                    async with axon.getLocalProxy() as proxy:
                        self.true(await proxy.has(asdfhash))



import synapse.lib.module as s_module

class AuthModule(s_module.CoreModule):

    def getModelDefs(self):

        modl = {
            'types': (
                ('auth:creds', ('guid', {}), {
                    'doc': 'A unique set of credentials used to access a resource.',
                }),
                ('auth:access', ('guid', {}), {
                    'doc': 'An instance of using creds to access a resource.',
                }),
            ),
            'forms': (
                ('auth:creds', {}, (
                    ('email', ('inet:email', {}), {
                        'doc': 'The email address used to identify the user.',
                    }),
                    ('user', ('inet:user', {}), {
                        'doc': 'The user name used to identify the user.',
                    }),
                    ('phone', ('tel:phone', {}), {
                        'doc': 'The phone number used to identify the user.',
                    }),
                    ('passwd', ('inet:passwd', {}), {
                        'doc': 'The password used to authenticate.',
                    }),
                    ('passwdhash', ('it:auth:passwdhash', {}), {
                        'doc': 'The password hash used to authenticate.',
                    }),
                    ('account', ('it:account', {}), {
                        'doc': 'The account that the creds allow access to.',
                    }),
                    ('website', ('inet:url', {}), {
                        'doc': 'The base URL of the website that the credentials allow access to.',
                    }),
                    ('host', ('it:host', {}), {
                        'doc': 'The host that the credentials allow access to.',
                    }),
                    ('wifi:ssid', ('inet:wifi:ssid', {}), {
                        'doc': 'The WiFi SSID that the credentials allow access to.',
                    }),
                    ('web:acct', ('inet:web:acct', {}), {
                        'doc': 'The web account that the credentials allow access to.',
                    }),
                    # TODO x509, rfid, mat:item locks/keys
                )),

                ('auth:access', {}, (
                    ('creds', ('auth:creds', {}), {
                        'doc': 'The credentials used to attempt access.',
                    }),
                    ('time', ('time', {}), {
                        'doc': 'The time of the access attempt.',
                    }),
                    ('success', ('bool', {}), {
                        'doc': 'Set to true if the access was successful.',
                    }),
                    ('person', ('ps:person', {}), {
                        'doc': 'The person who attempted access.',
                    }),
                )),
            ),
        }
        name = 'auth'
        return ((name, modl), )
